from . import DataManager
import pandas as pd
import dataclasses
import datetime
from fund.db.database import RawDatabaseConnector, BasicDatabaseConnector, DerivedDatabaseConnector
from ..struct import AssetWeight, AssetPrice, FundScoreParam, FundScoreHelper
from ..view.raw_models import *
from ..view.basic_models import *
from ..view.derived_models import *


@dataclasses.dataclass
class FundDataTables:
    # basic
    trading_days: pd.DataFrame = None
    index_info: pd.DataFrame = None
    index_price: pd.DataFrame = None
    index_pct: pd.DataFrame = None
    fund_info: pd.DataFrame = None
    fund_score: pd.DataFrame = None
    fund_list: set = None
    fund_nav: pd.DataFrame = None

    def __post_init__(self):
        self.fund_list = self.fund_list or set([])

    def __repr__(self):
        s = '<DTS'
        for k, v in self.__dict__.items():
            s += f' {k}:{v.shape if isinstance(v, pd.DataFrame) else (len(v) if isinstance(v, set) or isinstance(v, list) else None)}'
        s += '>'
        return s

class FundDataManager(DataManager):

    def __init__(self, start_time=None, end_time=None, fund_score_param: FundScoreParam=None):
        DataManager.__init__(self, 
            start_time or datetime.datetime(2005, 1, 1), 
            end_time or datetime.datetime.now()
        )
        self._fund_score_param = fund_score_param or FundScoreParam(tag_type=1, score_method=1, is_full=1)
        self.dts = FundDataTables()

    def init(self, index_list=None):
        index_list = index_list or list(AssetWeight.__dataclass_fields__.keys())
        # get derived data first
        with DerivedDatabaseConnector().managed_session() as derived_session:
            # fund_score
            tag_names = list(map(lambda x: FundScoreHelper.get(x), index_list))
            _fund_score_query = derived_session.query(
                    FundScore.fund_id,
                    FundScore.datetime,
                    FundScore.score, 
                    FundScore.tag_name, 
                    FundScore.score_method,
                    FundScore.is_full,
                    FundScore.tag_type
                ).filter(
                    FundScore.tag_name.in_(tag_names),
                    FundScore.score_method == self._fund_score_param.score_method,
                    FundScore.is_full == self._fund_score_param.is_full,
                    FundScore.tag_type == self._fund_score_param.tag_type
                )
            self.dts.fund_score = pd.read_sql(_fund_score_query.statement, _fund_score_query.session.bind)
            self.dts.fund_score['index_id'] = self.dts.fund_score.tag_name.apply(lambda x: FundScoreHelper.parse(x))
            # update fund_list with tag score
            self.dts.fund_list = set(self.dts.fund_score.fund_id)

        # info is necessary
        with BasicDatabaseConnector().managed_session() as quant_session:
            def fetch_table(view):
                query = quant_session.query(view)
                return pd.read_sql(query.statement, query.session.bind)
            self.dts.fund_info = fetch_table(FundInfo)
            self.dts.trading_days = fetch_table(TradingDayList)
            # index
            self.dts.index_info = fetch_table(IndexInfo)
            _index_query = quant_session.query(IndexPrice.index_id, IndexPrice.datetime, IndexPrice.close).filter(IndexPrice.index_id.in_(index_list))
            self.dts.index_price = pd.read_sql(_index_query.statement, _index_query.session.bind).pivot_table(index='datetime', columns='index_id', values='close').fillna(method= 'ffill')
            # fund nav
            _fund_nav_query = quant_session.query(
                    FundNav.fund_id,
                    FundNav.adjusted_net_value,
                    FundNav.subscribe_status, 
                    FundNav.redeem_status, 
                    FundNav.datetime
                ).filter(
                    FundNav.fund_id.in_(self.dts.fund_list),
                    FundNav.datetime >= self.start_date,
                    FundNav.datetime <= self.end_date,
                )
            self.dts.fund_nav = pd.read_sql(_fund_nav_query.statement, _fund_nav_query.session.bind)

        # raw index val pct, only used in tactic asset allocation
        with RawDatabaseConnector().managed_session() as quant_session:
            query = quant_session.query(IndexValPct)
            df = pd.read_sql(query.statement, query.session.bind).pivot_table(index='datetime', columns='index_id', values='pe_pct').fillna(method='ffill')
            self.dts.index_pct = df.rename(columns=FundScoreHelper.TACTIC_ASSET_CONVERT)
            '''
            self.index_pct = {}
            self.index_pct['hs300'] = index_valpct[index_valpct['index_id'] == '000300.XSHG'][['datetime', 'pe_pct']].set_index('datetime')
            self.index_pct['csi500'] = index_valpct[index_valpct['index_id'] == '000905.XSHG'][['datetime', 'pe_pct']].set_index('datetime')
            self.index_pct['hs300']  = self.index_data[['hs300']].join(self.index_pct['hs300']).fillna(method = 'ffill')[['pe_pct']]
            self.index_pct['csi500'] = self.index_data[['csi500']].join(self.index_pct['csi500']).fillna(method = 'ffill')[['pe_pct']]
            '''
        print(self.dts)

    def get_index_pcts(self, dt):
        # jch: only pct within 7 days take effect
        INDEX_PCT_EFFECTIVE_DELAY_DAY_NUM = 7
        res = {}
        for index_id in self.dts.index_pct.columns:
            df = self.dts.index_pct[index_id]
            _filtered = df[(df.index <= dt) & (df.index >= dt - datetime.timedelta(days=INDEX_PCT_EFFECTIVE_DELAY_DAY_NUM))]
            if len(_filtered) > 0:
                res[index_id] = _filtered.iloc[-1]
        return res

    def get_fund_score(self, dt, index_id) -> dict: 
        df = self.dts.fund_score
        # jch: only score within 14 days take effect
        SCORE_EFFECTIVE_DELAY_DAY_NUM = 14
        # 注意，对于dt天结束时的分析，可以选用当天的打分！
        # TODO: 打分这里，可以考虑根据时间最后打分时间到现在，进行分数的适当扣减？
        _filtered = df[(df.index_id==index_id) & (df.datetime <= dt) & (df.datetime >= dt - datetime.timedelta(days=SCORE_EFFECTIVE_DELAY_DAY_NUM))]
        return {fund_id: _filtered[_filtered.fund_id==fund_id].iloc[-1].score for fund_id in set(_filtered.fund_id)}

    def get_fund_info(self, fund_id=None):
        '''
        fund_id                                                   000001!0
        wind_id                                                  000001.OF
        transition                                                       0
        order_book_id                                               000001
        desc_name                                                     华夏成长
        start_date                                              2001-12-18
        end_date                                                2040-12-31
        wind_class_1                                                 混合型基金
        wind_class_2                                               偏股混合型基金
        manager_id       王亚伟(20011218-20050412)\r\n田擎(20040227-20051029...
        company_id                                                    华夏基金
        benchmark                                                        0
        full_name                                               华夏成长证券投资基金
        currency                                                       CNY
        base_fund_id                                                     0
        is_structured                                                    0
        is_open                                                          0
        asset_type                                                    null
        update_time                                             2020-03-24
        manage_fee                                                     1.5
        trustee_fee                                                   0.25
        purchase_fee                                                   NaN
        redeem_fee                                                     NaN
        note                                                          None
        track_index                                                   none
        '''
        if fund_id:
            df = self.dts.fund_info
            _filtered = df[df['fund_id']=='000001!0']
            return None if len(_filtered) == 0 else _filtered.iloc[0]
        else:
            return self.dts.fund_info.copy()

    def get_trading_days(self):
        return self.dts.trading_days.copy()

    def get_index_price(self, dt=None):
        if dt:
            return self.dts.index_price.loc[dt]
        else:
            return self.dts.index_price.copy()
    
    def get_index_price_data(self, dt):
        _index_price = self.get_index_price(dt)
        return AssetPrice(**_index_price.to_dict())

if __name__ == '__main__':
    m = FundDataManager('20100101', '20200101')
    m.init()
    print(m.get_fund_info(fund_id=None))
    print(m.get_fund_info(fund_id='000001!0'))
    print(m.get_trading_days())


from sqlalchemy import CHAR, Column, Integer, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import DOUBLE, DATE, TINYINT

Base = declarative_base()


class IndexInfo(Base):
    '''指数信息表'''

    __tablename__ = 'index_info'
    index_id = Column(CHAR(20), primary_key=True)

    order_book_id= Column(CHAR(20)) #米筐ID
    #web_id= Column(CHAR(20)) # 数据所在网页ID
    industry_tag= Column(CHAR(64)) # 行业标签
    tag_method= Column(CHAR(64)) # 估值评分采用方法
    desc_name= Column(CHAR(64)) # 名称

class StockInfo(Base):
    '''股票信息表'''

    __tablename__ = 'stock_info'
    stock_id = Column(CHAR(20), primary_key=True) # 股票ID
    rq_id = Column(CHAR(20)) # 米筐ID


class FundSize(Base):
    '''米筐基金最新规模'''
    
    __tablename__ = 'fund_size'

    fund_id = Column(CHAR(10), primary_key=True) # 基金id
    latest_size = Column(DOUBLE(asdecimal=False)) # 最新规模
    update_time = Column(DATE) # 更新日期


class FundInfo(Base):
    '''基金信息表'''

    __tablename__ = 'fund_info'
    fund_id = Column(CHAR(10),primary_key=True) # 基金ID
    
    wind_id = Column(CHAR(10)) # Wind基金ID
    transition = Column(Integer) # 基金变更次数
    order_book_id = Column(CHAR(10)) # RiceQuant基金ID
    desc_name = Column(CHAR(64)) # 基金名称
    start_date = Column(DATE) # 成立日期
    end_date = Column(DATE) # 关闭日期
    wind_class_1 = Column(CHAR(64)) # Wind基金类型
    wind_class_2 = Column(CHAR(64)) # Wind基金二级类型
    manager_id = Column(Text) # 基金经理
    company_id = Column(CHAR(64)) # 基金公司
    benchmark = Column(CHAR(255)) # 业绩基准
    full_name = Column(CHAR(255)) # 基金全名
    currency = Column(CHAR(20)) # 币种
    base_fund_id = Column(CHAR(20)) # 分级基金基础基金代号
    is_structured = Column(TINYINT(1)) # 是否为分级基金
    is_open = Column(TINYINT(1)) # 是否为开放式基金
    asset_type = Column(CHAR(32)) #资产类别
    update_time = Column(DATE) # 更新时间
    manage_fee = Column(DOUBLE(asdecimal=False)) # 管理费
    trustee_fee = Column(DOUBLE(asdecimal=False)) # 托管费
    purchase_fee = Column(DOUBLE(asdecimal=False)) # 申购费
    redeem_fee = Column(DOUBLE(asdecimal=False)) # 赎回费
    note = Column(CHAR(64)) # 附加信息
    track_index = Column(CHAR(20)) # 跟踪指数


class TradingDayList(Base):
    '''交易日列表'''

    __tablename__ = 'trading_day_list'
    datetime = Column(DATE, primary_key=True)


class FundNav(Base):
    '''基金净值表'''

    __tablename__ = 'fund_nav'
    fund_id = Column(CHAR(20), primary_key=True) # 合约代码
    datetime = Column(DATE, primary_key=True) # 日期

    unit_net_value = Column(DOUBLE(asdecimal=False)) # 单位净值
    acc_net_value = Column(DOUBLE(asdecimal=False)) # 累积单位净值
    adjusted_net_value = Column(DOUBLE(asdecimal=False)) # 复权净值
    change_rate = Column(DOUBLE(asdecimal=False)) # 涨跌幅
    daily_profit = Column(DOUBLE(asdecimal=False)) # 每万元收益（日结型货币基金专用）
    weekly_yield = Column(DOUBLE(asdecimal=False)) # 7日年化收益率（日结型货币基金专用）
    redeem_status = Column(Integer) # 赎回状态，开放 - Open, 暂停 - Suspended, 限制大额申赎 - Limited, 封闭期 - Close
    subscribe_status = Column(Integer) # 订阅状态，开放 - Open, 暂停 - Suspended, 限制大额申赎 - Limited, 封闭期 - Close
    order_book_id = Column(CHAR(10)) # RiceQuant基金ID
    

class IndexPrice(Base):
    '''指数价格表'''

    __tablename__ = 'index_price'
    id = Column(Integer, primary_key=True)
    # TODO: (index_id, datetime) as primary_key
    index_id = Column(CHAR(20)) # 指数id
    datetime = Column(DATE) # 日期
    volume = Column(DOUBLE(asdecimal=False)) # 交易量
    low = Column(DOUBLE(asdecimal=False)) # 最低价
    close = Column(DOUBLE(asdecimal=False)) # 收盘价
    high = Column(DOUBLE(asdecimal=False)) # 最高价
    open = Column(DOUBLE(asdecimal=False)) # 开盘价
    total_turnover = Column(DOUBLE(asdecimal=False)) # 成交额
    ret = Column(DOUBLE(asdecimal=False)) # 收益率


class StockPrice(Base):
    '''股票价格表'''

    __tablename__ = 'stock_price'
    id = Column(Integer, primary_key=True)
    # TODO: (stock_id, datetime) as primary_key
    stock_id = Column(CHAR(20)) # 指数id
    datetime = Column(DATE) # 日期
    open = Column(DOUBLE(asdecimal=False)) # 开盘价
    close = Column(DOUBLE(asdecimal=False)) # 收盘价
    high = Column(DOUBLE(asdecimal=False)) # 最高价
    low = Column(DOUBLE(asdecimal=False)) # 最低价
    limit_up = Column(DOUBLE(asdecimal=False)) # 涨停价
    limit_down = Column(DOUBLE(asdecimal=False)) # 跌停价
    total_turnover = Column(DOUBLE(asdecimal=False)) # 交易额
    volume = Column(DOUBLE(asdecimal=False)) # 交易量  
    num_trades = Column(DOUBLE(asdecimal=False)) # 交易笔数
    adj_close = Column(DOUBLE(asdecimal=False)) # 后复权价格
    post_adj_factor = Column(DOUBLE(asdecimal=False)) # 后复权因子
    ret = Column(DOUBLE(asdecimal=False)) # 日收益


class FundRet(Base):
    '''基金历史收益'''

    __tablename__ = 'fund_ret'
    id = Column(Integer, primary_key=True)
    # TODO: (fund_id, datetime) as primary_key
    fund_id = Column(CHAR(10)) # 原始基金ID
    datetime = Column(DATE) # 日期
    w1_ret = Column(DOUBLE(asdecimal=False)) # 近一周收益率
    m1_ret = Column(DOUBLE(asdecimal=False)) # 近一月收益率
    m3_ret = Column(DOUBLE(asdecimal=False)) # 近一季度收益率
    m6_ret = Column(DOUBLE(asdecimal=False)) # 近半年收益率
    y1_ret = Column(DOUBLE(asdecimal=False)) # 近一年收益率
    mdd = Column(DOUBLE(asdecimal=False)) # 最大回撤（成立以来）
    annual_ret = Column(DOUBLE(asdecimal=False)) # 年化收益（成立以来）
    avg_size = Column(DOUBLE(asdecimal=False)) # 平均规模（成立以来）
    sharpe_ratio = Column(DOUBLE(asdecimal=False)) # 夏普率（成立以来）
    info_ratio = Column(DOUBLE(asdecimal=False)) # 夏普率（成立以来）


class FundRatingLatest(Base):
    '''基金最新评级'''
    

    __tablename__ = 'fund_rating_latest'

    fund_id = Column(CHAR(10), primary_key=True)  # 基金id
    zs = Column(DOUBLE(asdecimal=False))  # 招商评级
    sh3 = Column(DOUBLE(asdecimal=False))  # 上海证券评级三年期
    sh5 = Column(DOUBLE(asdecimal=False))  # 上海证券评级五年期
    jajx = Column(DOUBLE(asdecimal=False))  # 济安金信评级
    update_time = Column(DATE)  # 更新日期





import pandas as pd
from python.fund.db.database import RawDatabaseConnector, BasicDatabaseConnector, DerivedDatabaseConnector
from python.fund.db.view_models import FundDailyCollection
from python.fund.db.raw_models import RqFundIndicator
from python.fund.db.basic_models import FundNav, FundInfo, FundRatingLatest, FundSize
from python.fund.db.derived_models import FundAlpha, FundScore, FundIndicator
from sqlalchemy import create_engine
from api_investment.constants import ExchangeStatus


def get_fund_df():
    with BasicDatabaseConnector().managed_session() as mn_session:
        try:
            query = mn_session.query(
                FundInfo.fund_id,
                FundInfo.order_book_id,
                FundInfo.wind_class_1,
                FundInfo.start_date,
                FundInfo.manager_id,
                FundInfo.company_id,
                FundInfo.benchmark,
                FundInfo.desc_name,
                FundInfo.track_index,
            )
            funds = pd.read_sql(query.statement, query.session.bind)
            return funds.set_index('fund_id')
        except Exception as e:
            print('Failed fill get data <err_msg> {}'.format(e))

def get_nav_df():
    with BasicDatabaseConnector().managed_session() as mn_session:
        try:
            latest_time = mn_session.query(FundNav).order_by(FundNav.datetime.desc()).limit(1).one_or_none()
            latest_time = latest_time.datetime
            query = mn_session.query(
                FundNav.fund_id,
                FundNav.unit_net_value,
                FundNav.acc_net_value,
                FundNav.adjusted_net_value,
                FundNav.datetime,
                FundNav.redeem_status,
                FundNav.subscribe_status,
            ).filter(FundNav.datetime==latest_time)

            nav = pd.read_sql(query.statement, query.session.bind)
            return nav.set_index('fund_id'), latest_time
        except Exception as e:
            print('Failed fill get data <err_msg> {}'.format(e))

def get_indicator_df():
    with RawDatabaseConnector().managed_session() as quant_session:
        try:
            latest_time = quant_session.query(RqFundIndicator).order_by(RqFundIndicator.datetime.desc()).limit(1).one_or_none()
            latest_time = latest_time.datetime
            query = quant_session.query(
                RqFundIndicator.order_book_id,
                RqFundIndicator.annualized_returns,
                RqFundIndicator.annualized_risk,
                RqFundIndicator.average_size,
                RqFundIndicator.information_ratio,
                RqFundIndicator.last_month_return,
                RqFundIndicator.last_six_month_return,
                RqFundIndicator.last_three_month_return,
                RqFundIndicator.last_twelve_month_return,
                RqFundIndicator.last_week_return,
                RqFundIndicator.year_to_date_return,
                RqFundIndicator.to_date_return,
                RqFundIndicator.sharp_ratio,
                RqFundIndicator.max_drop_down,
            ).filter(RqFundIndicator.datetime==latest_time)
            indicator = pd.read_sql(query.statement, query.session.bind)
            return indicator

        except Exception as e:
            print('Failed fill get data <err_msg> {}'.format(e))

def get_institution_rating_df():
    with BasicDatabaseConnector().managed_session() as mn_session:
        try:
            latest_time = mn_session.query(FundRatingLatest).order_by(FundRatingLatest.update_time.desc()).limit(1).one_or_none()
            latest_time = latest_time.update_time
            print(latest_time)
            query = mn_session.query(
                FundRatingLatest.fund_id,
                FundRatingLatest.zs,
                FundRatingLatest.sh3,
                FundRatingLatest.sh5,
                FundRatingLatest.jajx,
            ).filter(FundRatingLatest.update_time==latest_time)
            rating = pd.read_sql(query.statement, query.session.bind)
            return rating

        except Exception as e:
            print('Failed fill get data <err_msg> {}'.format(e))

def get_size_df():
    with BasicDatabaseConnector().managed_session() as mn_session:
        try:
            query = mn_session.query(
                FundSize.fund_id,
                FundSize.latest_size
            )
            df = pd.read_sql(query.statement, query.session.bind)
            df = df.drop_duplicates(subset=['fund_id'], keep='last')
            df = df.set_index('fund_id')
            return df
        except Exception as e:
            print('Failed fill get data <err_msg> {}'.format(e))

def get_alpha_df():
    with DerivedDatabaseConnector().managed_session() as mn_session:
        try:
            latest_time = mn_session.query(FundAlpha).order_by(FundAlpha.datetime.desc()).limit(1).one_or_none()
            latest_time = latest_time.datetime
            print(latest_time)
            query = mn_session.query(
                FundAlpha.fund_id,
                FundAlpha.track_err,
                FundAlpha.this_y_alpha,
                FundAlpha.cumulative_alpha,
                FundAlpha.w1_alpha,
                FundAlpha.m1_alpha,
                FundAlpha.m3_alpha,
                FundAlpha.m6_alpha,
                FundAlpha.y1_alpha,
                FundAlpha.y3_alpha,
                FundAlpha.y5_alpha,
                FundAlpha.y10_alpha,
            ).filter(FundAlpha.datetime==latest_time)
            df = pd.read_sql(query.statement, query.session.bind)
            df = df.drop_duplicates(subset=['fund_id'], keep='last')
            return df

        except Exception as e:
            print('Failed fill get data <err_msg> {}'.format(e))

def get_score_df():
    with DerivedDatabaseConnector().managed_session() as mn_session:
        try:
            latest_time = mn_session.query(FundScore.datetime).order_by(FundScore.datetime.desc()).limit(1).one_or_none()[0]
            print(latest_time)
            query = mn_session.query(
                FundScore.fund_id,
                FundScore.tag_name,
                FundScore.score,
            ).filter(FundScore.datetime==latest_time)
            df = pd.read_sql(query.statement, query.session.bind)
            df = df.drop_duplicates(subset=['fund_id'], keep='last')
            return df

        except Exception as e:
            print('Failed get_score_df <err_msg> {}'.format(e))

def get_tag_fund_indicator_df():
    with DerivedDatabaseConnector().managed_session() as mn_session:
        try:
            latest_time = mn_session.query(FundIndicator).order_by(FundIndicator.datetime.desc()).limit(1).one_or_none()
            latest_time = latest_time.datetime
            print(latest_time)
            query = mn_session.query(
                FundIndicator.fund_id,
                FundIndicator.alpha,
                FundIndicator.beta,
                FundIndicator.track_err,
                FundIndicator.fee_rate,
            ).filter(FundIndicator.datetime==latest_time)
            df = pd.read_sql(query.statement, query.session.bind)
            df = df.rename(columns={'track_err': 'tag_track_err'})
            df = df.drop_duplicates(subset=['fund_id'], keep='last')
            return df

        except Exception as e:
            print('Failed get_fund_indicator_df <err_msg> {}'.format(e))

def append_data(table_name, data_append_directly_data_df):
    engine = create_engine(
            'mysql+pymysql://wangjun:Wangjun123456@fundinfo.cq1tbd5lkqzo.rds.cn-northwest-1.amazonaws.com.cn:3306/view')

    if not data_append_directly_data_df.empty:
        data_append_directly_data_df.to_sql(table_name, engine, index = False, if_exists = 'append')
        print('新数据已插入')
    else:
        print('没有需要插入的新数据')

def get_exchange_status(x):
    if x['redeem_status'] == ExchangeStatus.LIMITED:
        x['redeem_status'] = ExchangeStatus.OPEN
    if x['subscribe_status'] == ExchangeStatus.LIMITED:
        x['subscribe_status'] = ExchangeStatus.OPEN
    if x['redeem_status'] == ExchangeStatus.CLOSE and x['subscribe_status'] == ExchangeStatus.CLOSE:
        return '暂停交易'
    if x['redeem_status'] == ExchangeStatus.SUSPENDED:
        return '暂停赎回'
    if x['subscribe_status'] == ExchangeStatus.SUSPENDED:
        return '暂停申购'
    if x['redeem_status'] == ExchangeStatus.OPEN and x['subscribe_status'] == ExchangeStatus.OPEN:
        return '正常开放'
    return '暂无状态'

def get_found_to_now(x):
    if pd.isna(x):
        return None
    return round((pd.datetime.now().date() - x).days / 365, 2)

def modify_fund_manager(x):
    try:
        return x.split('\r\n')[-1].split('(')[0]
    except:
        return x

def compute_daily_collection():
    funds = get_fund_df()
    funds = funds.reset_index()
    funds = funds.drop_duplicates(subset=['fund_id'], keep='last')
    funds = funds.set_index('fund_id')
    print('info', funds)
    print('-' * 100)

    nav, latest_time = get_nav_df()
    print(latest_time)
    nav = nav.reset_index()
    nav = nav.drop_duplicates(subset=['fund_id'], keep='last')
    nav = nav.set_index('fund_id')
    print('nav', nav)
    print('-' * 100)

    alpha = get_alpha_df()
    alpha = alpha.set_index('fund_id')
    print(alpha)
    print('-' * 100)
    size = get_size_df()
    print(size)
    print('-' * 100)
    score = get_score_df().set_index('fund_id')
    print(score)
    print('-' * 100)
    tag_indicator = get_tag_fund_indicator_df().set_index('fund_id')
    print(tag_indicator)
    print('-' * 100)

    print('处理机构评级')
    rating = get_institution_rating_df()
    rating = rating.drop_duplicates(subset=['fund_id'], keep='last')
    rating = rating.set_index('fund_id')
    print(rating)
    print('-' * 100)

    print('处理净值')
    df = pd.concat([funds, nav, alpha, size, rating, score, tag_indicator], axis=1, sort=False)
    df.index.name = 'fund_id'
    print('_'*100)
    print(df)

    print('处理行情')
    indicator = get_indicator_df()
    indicator = indicator.dropna(subset=['order_book_id'])
    indicator['fund_id'] = indicator['order_book_id'] + '!0'
    indicator = indicator.drop_duplicates(subset=['order_book_id'], keep='last')
    indicator = indicator.drop(['order_book_id'], axis=1)
    indicator = indicator.set_index('fund_id')
    print(indicator)
    print('-' * 100)

    df = pd.concat([df, indicator], axis=1, sort=False, join_axes=[df.index])
    df.index.name = 'fund_id'
    df = df.reset_index()
    df = df.dropna(subset=['fund_id'])
    df['exchange_status'] = df.apply(get_exchange_status, axis=1)
    df['found_to_now'] = df['start_date'].apply(get_found_to_now)
    df['manager_id'] = df['manager_id'].apply(modify_fund_manager)
    df = df.rename(columns={
        'wind_class_1': 'wind_class_I',
        'start_date': 'found_date',
        'manager_id': 'fund_manager',
        'company_id': 'company_name',
        'desc_name': 'symbol',
    })
    df = df.drop(['redeem_status', 'subscribe_status'], axis=1)
    df = df.sort_values(['fund_id'])
    df = df.drop_duplicates(subset=['order_book_id'], keep='first')
    df = df.dropna(subset=['order_book_id'])

    # df.to_csv('./hhh.csv')
    append_data(FundDailyCollection.__tablename__, df)

    print(df)


if __name__ == '__main__':
    compute_daily_collection()





from ...data.struct import AssetWeight, AssetPrice, AssetPosition, AssetValue
from ...data.struct import TAAParam
from . import Helper


class SAAHelper(Helper):

    def __init__(self):
        pass

    def setup(self, saa: AssetWeight):
        self.saa = saa

    def on_price(self, dt, asset_price: AssetPrice):
        cur_saa = self.saa.copy()
        for k, v in asset_price.__dict__.items():
            if asset_price.isnan(v):
                cur_saa.__dict__[k] = 0
        cur_saa.rebalance()
        return cur_saa


class TAAStatusMode:
    NORMAL = 'normal'
    IN_LOW = 'low'
    IN_HIGH = 'high'

class TAAHelper(Helper):

    def __init__(self, taa_params: TAAParam=None):
        self.params = taa_params or TAAParam()
        self.tactic_status = {}

    def on_price(self, dt, asset_price: AssetPrice, cur_saa: AssetWeight, asset_pct: dict):
        taa = cur_saa.copy()
        taa_effected = False
        mode_changed = False
        for index_id, target_w in taa.__dict__.items():

            if index_id in asset_pct:
                val_pct = asset_pct[index_id]
                cur_mode = self.tactic_status.get(index_id, TAAStatusMode.NORMAL)
                new_mode = TAAStatusMode.NORMAL
                tactic_w = target_w

                if cur_mode == TAAStatusMode.NORMAL:

                    if val_pct >= self.params.HighThreshold:
                        tactic_w = max(target_w - self.params.HighMinus, 0)
                        new_mode = TAAStatusMode.IN_HIGH
                    elif val_pct <= self.params.LowThreshold:
                        tactic_w = min(target_w + self.params.LowPlus, 1)
                        new_mode = TAAStatusMode.IN_LOW
                    
                elif cur_mode == TAAStatusMode.IN_LOW:

                    if val_pct < self.params.LowStop:
                        tactic_w = min(target_w + self.params.LowPlus, 1)
                        new_mode = TAAStatusMode.IN_LOW

                elif cur_mode == TAAStatusMode.IN_HIGH:

                    if val_pct > self.params.HighStop:
                        tactic_w = max(target_w - self.params.HighMinus, 0)
                        new_mode = TAAStatusMode.IN_HIGH
                    
                else:
                    assert False, 'should not be here!'

                self.tactic_status[index_id] = new_mode
                if new_mode != TAAStatusMode.NORMAL:
                    taa.__dict__[index_id] = tactic_w

                mode_changed = mode_changed or new_mode != cur_mode
                taa_effected = taa_effected or new_mode != TAAStatusMode.NORMAL
        taa.rebalance()
        if taa_effected:
            print(f'taa: {dt} (mode){mode_changed} (taa){taa_effected} (saa){cur_saa} (taa){taa}')
        return taa

    


import datetime
import json
import threading
import time
import uuid
import redis

class RedisClient(object):

    def __init__(self, redis_setting):
        self.redis_client = redis.StrictRedis(host=redis_setting.host,
                                              port=redis_setting.port,
                                              db=redis_setting.db,
                                              password=redis_setting.password,
                                              decode_responses=True)

        self._is_connected = self.ping_redis()
        if not self.is_connected():
            raise RuntimeError('redis not connected')

        print('[redis] current connection status (conn){}'.format(self._is_connected))

    def is_connected(self):
        return self._is_connected

    def ping_redis(self):
        try:
            self.redis_client.ping()

        except redis.exceptions.ConnectionError as e:
            self._is_connected = False
            print('[ping] redis connection failed! (err){}'.format(e))
            return False

        else:
            self._is_connected = True
            return True

    def disconnect(self):
        self._is_connected = False


class WechatMessageRedisClient(RedisClient):

    def __init__(self, redis_setting):
        
        RedisClient.__init__(self, redis_setting)

        self.key = redis_setting.key

    def get_message(self):
        message = self.redis_client.lpop(self.key)
        return message



class WechatMessageParser(object):

    def parse(self, msg_from, msg_to, msg_content):
        result = {}
        result['scenario_id'] = 1
        result['param'] = {'fund_id': msg_content}
        return result


#!/usr/bin/env python
# coding: utf-8
import numpy as np
import pandas as pd
import rqdatac as rq
from .EMQuantAPI_Python.python3.EmQuantAPI import *
import datetime
import json
import csv
from ..db.database import RawDatabaseConnector
from ..db.raw_models import *
from ..util.constant import STOCK_VALUATION_FACTORS, FUND_INDICATORS
from ..data_api.basic_new import BasicDataApi

class EmRawDataDownloader(object):

    def __init__(self):
        # ForceLogin
        # 取值0，当线上已存在该账户时，不强制登录
        # 取值1，当线上已存在该账户时，强制登录，将前一位在线用户踢下线
        options = "ForceLogin=1"
        loginResult = c.start(options, mainCallBack=self._main_callback)
        if(loginResult.ErrorCode != 0):
            print("EM login failed")
            exit()

        self._fund_ids = self._get_fund_ids()

    def _upload_raw(self, df, table_name, if_exists='append'):
        print(table_name)
        print(df)
        # df.to_sql(table_name, RawDatabaseConnector().get_engine(), index=False, if_exists=if_exists)

    def _get_fund_ids(self):
        fund_id_list = []
        with open('tag.csv') as csvfile:
            csvreader = csv.reader(csvfile)
            for row in csvreader:
                fund_id_list.append(f'{row[1][:-2]}.OF')
        print(fund_id_list)

        fund_id_str = ','.join(fund_id_list)
        return fund_id_str

    def _main_callback(self, quantdata):
        """
        mainCallback 是主回调函数，可捕捉连接错误
        该函数只有一个为c.EmQuantData类型的参数quantdata
        :param quantdata: c.EmQuantData
        """
        print (f"MainCallback: {quantdata}")

    def em_fund_nav(self, start_date, end_date):
        # TODO: compare date param with actual start_data, end_data of each fund, then get data
        df = c.csd(self._fund_ids, "ADJUSTEDNAV", start_date, end_date, "AdjustFlag=2,Ispandas=1")

        # self._upload_raw(df_all, EmFundNav.__table__.name)

    def em_index(self, start_date, end_date):
        # ----------------------
        # H11073.CSI,信用债
        # SPX.GI,标普500
        # GDAXI.GI,德国DAX
        # N225.GI,日经225
        # 000012.SH,利率债
        # 000300.SH,沪深300
        # 000905.SH,中证500
        # 399006.SZ,创业板
        # 801181.SWI,房地产开发申万
        # ----------------------
        # USDCNY.IB,美元兑人民币市询价,TODO: '''ErrorCode=10001012, ErrorMsg=insufficient user access, Data={}'''
        # JPYCNY.IB,日元兑人民币市询价,TODO: '''ErrorCode=10001012, ErrorMsg=insufficient user access, Data={}'''
        # EURCNY.IB,欧元兑人民币市询价,TODO: '''ErrorCode=10001012, ErrorMsg=insufficient user access, Data={}'''
        # ----------------------
        # AU0.SHF,沪金主力连续
        # BC.ICE,布伦特原油当月连续,TODO: always get None
        # ----------------------
        codes = ("H11073.CSI,SPX.GI,GDAXI.GI,N225.GI,000012.SH,000300.SH,000905.SH,399006.SZ,801181.SWI,"
            "USDCNY.IB,JPYCNY.IB,EURCNY.IB,"
            "AU0.SHF,BC.ICE")
        df = c.csd(codes, "CLOSE", start_date, end_date, "Ispandas=1")

        # self._upload_raw(df, EmIndexPrice.__table__.name)

    def download_all(self, start_date, end_date):
        self.em_fund_nav(start_date, end_date)
        # self.em_index(start_date, end_date)




import pymysql
import pandas as pd
import numpy as np
from pprint import pprint
import sys

'''
Load data from database and transhfer data to dataframe
'''


class LoadDataFromDB():

    def __init__(self, user_name, password, url='fundinfo.cq1tbd5lkqzo.rds.cn-northwest-1.amazonaws.com.cn',
                 database_name='quant_data', encode='utf8'):
        self.url = url
        self.user_name = user_name
        self.password = password
        self.database_name = database_name
        self.encode = encode
        try:
            self.connect = pymysql.connect(
                                host = self.url, 
                                user = self.user_name,
                                password = self.password,
                                database = self.database_name,
                                charset = self.encode 
                                )
            # print('成功链接数据库 %s'%(self.database_name))
            
        except:
            print('连接数据库失败')
        
    def construct_sql(self, table_name, column_names = None, select_columns = None, aim_values = None, operator = '=', mode = 'and', count = None):
        if column_names != None:
            column_names = ["`" + i + "`" for i in column_names]
        if select_columns != None:
            select_columns = ["`" + i + "`" for i in select_columns]
        table_name = self.database_name + '.' + table_name
        sql = ''
        if column_names:
            sql = 'select %s from %s' % (''.join([(i + ',') for i in column_names])[:-1], table_name)
            if select_columns:
                if len(select_columns) == 1:
                    sql = 'select %s from %s where %s%s%s'%(''.join([(i + ',') for i in column_names])[:-1], table_name, select_columns[0], operator[0], "'" + aim_values[0] + "'")
                elif len(select_columns) >1:
                    if isinstance(operator, str):
                        sql = 'select %s from %s where %s%s%s'%(''.join([(i + ',') for i in column_names])[:-1], table_name, select_columns[0], operator, "'" + aim_values[0] + "'")
                    
                        if isinstance(mode, str):
                            for i in range(1, len(select_columns)):
                                sql += ' %s %s%s%s'%(mode, select_columns[i], operator, "'" + aim_values[i] + "'")
                        elif isinstance(mode, list):
                            for i in range(1, len(select_columns)):
                                sql += ' %s %s%s%s'%(mode[i - 1], select_columns[i], operator, "'" + aim_values[i] + "'")
                            sql = sql[:-1]
                    elif isinstance(operator, list):
                        sql = 'select %s from %s where %s%s%s'%(''.join([(i + ',') for i in column_names])[:-1], table_name, select_columns[0], operator[0], "'" + aim_values[0] + "'")
                        if isinstance(mode, str):
                            for i in range(1, len(select_columns)):
                                sql += ' %s %s%s%s'%(mode, select_columns[i], operator[i], "'" + aim_values[i] + "'")
                        elif isinstance(mode, list):
                            for i in range(1, len(select_columns)):
                                sql += ' %s %s%s%s'%(mode[i - 1], select_columns[i], operator[i], "'" + aim_values[i] + "'")
            elif select_columns == None:
                pass
        elif column_names == None:
            if select_columns == None:
                sql = 'select * from %s' % (table_name)
            if select_columns:
                if len(select_columns) == 1:
                    sql = 'select * from %s where %s%s%s'%(table_name, select_columns[0], operator[0], "'" + aim_values[0] + "'")
                elif len(select_columns) >1:
                    if isinstance(operator, str):
                        sql = 'select * from %s where %s%s%s'%(table_name, select_columns[0], operator, "'" + aim_values[0] + "'")
                    
                        if isinstance(mode, str):
                            for i in range(1, len(select_columns)):
                                sql += ' %s %s%s%s'%(mode, select_columns[i], operator, "'" + aim_values[i] + "'")
                        elif isinstance(mode, list):
                            for i in range(1, len(select_columns)):
                                sql += ' %s %s%s%s'%(mode[i-1], select_columns[i], operator, "'" + aim_values[i] + "'")
                    elif isinstance(operator, list):
                        sql = 'select * from %s where %s%s%s'%(table_name, select_columns[0], operator[0], "'" + aim_values[0] + "'")
                        if isinstance(mode, str):
                            for i in range(1, len(select_columns)):
                                sql += ' %s %s%s%s'%(mode, select_columns[i], operator[i], "'" + aim_values[i] + "'")
                        elif isinstance(mode, list):
                            for i in range(1, len(select_columns)):
                                sql += ' %s %s%s%s'%(mode[i - 1], select_columns[i], operator[i], "'" + aim_values[i] + "'")
        if count == None:
            pass
        else:
            sql += ' limit %s' % (count)
        return sql

    def get_data_columns(self, cursor):
        return [i[0] for i in cursor.description]

    def transfer_data_to_df(self, data_tuple, cursor):
        columns = self.get_data_columns(cursor)
        data_df = pd.DataFrame(np.array([list(i) for i in data_tuple]))
        if data_df.shape[0] >0 and data_df.shape[1] > 0:
            data_df.columns = columns
            return data_df
        else:
            return None
                               
    def get_data(self, table_name, column_names = None, select_columns = None, aim_values = None, operator = '=', mode = 'and', count = None):
        sql = self.construct_sql(table_name, column_names, select_columns, aim_values, operator, mode, count)
        with self.connect.cursor() as cu:
            cu.execute(sql)
            result = cu.fetchall()
            return self.transfer_data_to_df(result, cu)
    
    def get_data_from_sql(self, sql):
        with self.connect.cursor() as cu:
            cu.execute(sql)
            result = cu.fetchall()
            return self.transfer_data_to_df(result, cu)


class DataLoader(LoadDataFromDB):

    def __init__(self, user_name, password, url='fundinfo.cq1tbd5lkqzo.rds.cn-northwest-1.amazonaws.com.cn',
                 database_name='quant_data', encode='utf8'):
        super(DataLoader, self).__init__(user_name, password, url, database_name, encode)
        # self.mappin_function()

    def get_all_table_names(self):
        with self.connect.cursor() as cu:
            cu.execute('show tables')
            table_name_list = cu.fetchall()
        return [i[0] for i in table_name_list]

    def mappin_function(self):
        all_table_names = self.get_all_table_names()
        for table_name in all_table_names:
            tmp = "self.get_%s = self.defind('%s', depend_date_columns = '%s')" % (table_name, table_name, 'datetime')
            exec(tmp)

    def defind(self, table_name, depend_date_columns=None):
        function = None
        if depend_date_columns:
            def data_load(start_time=None, end_time=None, count=10):
                if start_time and end_time:
                    return self.get_data(table_name=table_name, count=count,
                                         select_columns=[depend_date_columns, depend_date_columns],
                                         aim_values=[start_time, end_time], operator=['>=', '<='])
                else:
                    return self.get_data(table_name, count)

            function = data_load
            return function
        else:
            def data_load(count=10):
                return self.get_data(table_name, count)

            function = data_load
            return function






from collections import OrderedDict
from fund.util.logger import SurfingLogger

class DpModule(object):
    '''
    数据处理组件，执行具体计算任务
    （输入数据）-> DpModule -> （输出数据）
    '''

    def __init__(self):
        '''
        初始化，定义计算必须的输入数据字典项
        '''
        self._necessary_keys = []
        self._logger = SurfingLogger.get_logger(type(self).__name__)

    def run(self, data_in):
        '''
        数据处理主函数，供外部调用
        此函数调用内部计算流程
        :param data_in: 输入数据（字典，字典 key 为数据名称，value 为数据体）
        :return: 如果成功，输出数据字典（字典 key 为数据名称，value 为数据体）；如果失败输出 None
        '''
        if not self._verify_data_in(data_in):
            return None

        try:
            self._logger.info(f'{type(self).__name__} starts to run')
            return self._run(data_in)
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            self._logger.error(f'{type(self).__name__} run failed, error: {traceback.format_exc()}')
            return None

    def _verify_data_in(self, data_in):
        '''
        验证输入数据是否满足需求
        :param data_in: 输入数据（字典，字典 key 为数据名称，value 为数据体）
        :return Bool: True 表示满足，False 表示不满足
        '''
        for necessary_key in self._necessary_keys:
            if necessary_key not in data_in:
                self._logger.error(f'Missing necessary key {necessary_key} in PositionProcessor data_in, processor stopped')
                return False
        return True

    # TO BE OVERRIDDEN
    def _run(self, data_in):
        '''
        数据处理逻辑函数
        :param data_in: 输入数据（字典，字典 key 为数据名称，value 为数据体）
        :return: 如果成功，输出数据字典（字典 key 为数据名称，value 为数据体）；如果失败输出 None
        '''
        self._logger.error('_run(self, data_in) in DpModule is not overridden')
        return None

class DataProcessor(object):
    '''
    数据处理调度器，调度 DpModule 执行计算逻辑
    （输入数据）-> DataProcessor -> （输出数据）
    '''

    def __init__(self):
        '''
        初始化带有优先级的数据处理组件列表
        优先级数字越小代表优先级越高，同一优先级下组件无相互依赖，低优先级组件可依赖高优先级组件
        modules: {priority(int)->[DpModule]}
        '''
        self._logger = SurfingLogger.get_logger(type(self).__name__)
        
        self.modules = {}
        self.priorities = []
    
    def attach(self, priority, dp_module):
        if priority not in self.modules:
            self.modules[priority] = []
            self.priorities.append(priority)
            self.priorities.sort()

        self.modules[priority].append(dp_module)

    def process(self, data, **kwargs):
        '''
        数据处理主函数，供外部调用
        此函数调用内部计算流程
        :param data: 输入数据（字典，字典 key 为数据名称，value 为数据体）
        :return: True or False
        '''
        for priority in self.priorities:
            to_break = False
            for module in self.modules[priority]:
                module_output = module.run(data)
                if module_output is not None:
                    # TODO
                    data.update(module_output)
                else:
                    self._logger.error(f'Module {type(module).__name__} failed in processing')
                    to_break = True
            if to_break:
                return False
        return True


import inspect

class ReadableEnum(object):
    _code_name_cache = None

    @classmethod
    def init_cache(cls):
        cls._code_name_cache = ({}, {})
        for name in dir(cls):
            attr = getattr(cls, name)
            if name.startswith('_') or inspect.ismethod(attr):
                continue
            # assert attr not in cls._code_name_cache[0], f'failed to register {name}, {attr} is a registed value'
            # assert name not in cls._code_name_cache[1], f'failed to register {attr}, {name} is a registed enum'
            cls._code_name_cache[0][attr] = name
            cls._code_name_cache[1][name] = attr

    @classmethod
    def read(cls, code):
        if cls._code_name_cache is None:
            cls.init_cache()
        return cls._code_name_cache[0].get(code, code)

    @classmethod
    def parse(cls, code):
        if cls._code_name_cache is None:
            cls.init_cache()
        return cls._code_name_cache[1].get(code, code)


class BarraFactorType(ReadableEnum):
    BETA = 0 # 风险指数
    LEVERAGE = 1 # 杠杆率
    LONG_MOMENTUN = 2 # 长期动量
    QUALITY = 3 # 盈利能力
    SHORT_MOMENTUM = 4 # 短期动量
    SIZE = 5 # 市值
    TURNOVER = 6 # 换手率
    VALUE = 7 # 估值
    VOLATILITY = 8 # 收益波动率

STOCK_VALUATION_FACTORS = [
    'pe_ratio_lyr',
    'pe_ratio_ttm',
    'ep_ratio_lyr',
    'ep_ratio_ttm',
    'pcf_ratio_total_lyr',
    'pcf_ratio_total_ttm',
    'pcf_ratio_lyr',
    'pcf_ratio_ttm',
    'cfp_ratio_lyr',
    'cfp_ratio_ttm',
    'pb_ratio_lyr',
    'pb_ratio_ttm',
    'pb_ratio_lf',
    'book_to_market_ratio_lyr',
    'book_to_market_ratio_ttm',
    'book_to_market_ratio_lf',
    'dividend_yield_ttm',
    'peg_ratio_lyr',
    'peg_ratio_ttm',
    'ps_ratio_lyr',
    'ps_ratio_ttm',
    'sp_ratio_lyr',
    'sp_ratio_ttm',
    'market_cap',
    'market_cap_2',
    'market_cap_3',
    'a_share_market_val',
    'a_share_market_val_in_circulation',
    'ev_lyr',
    'ev_ttm',
    'ev_lf',
    'ev_no_cash_lyr',
    'ev_no_cash_ttm',
    'ev_no_cash_lf',
    'ev_to_ebitda_lyr',
    'ev_to_ebitda_ttm',
    'ev_no_cash_to_ebit_lyr',
    'ev_no_cash_to_ebit_ttm'
]

FUND_INDICATORS = [
    'last_week_return',
    'last_month_return',
    'last_three_month_return',
    'last_six_month_return',
    'last_twelve_month_return',
    'year_to_date_return',
    'to_date_return',
    'average_size',
    'annualized_returns',
    'annualized_risk',
    'sharp_ratio',
    'max_drop_down',
    'information_ratio'
]



from fund.db.models import (IndexPrice,StockInfo,StockValuation,StockFinancialIndicator,StockLatestNetProfit,IndustryPE)
from fund.util.timer import Timer
from fund.util.logger import SurfingLogger
from fund.util.utility import Utility
from .database import QuantDatabaseConnector,DerivedDatabaseConnector
from .common import DataType

import json
import pandas as pd
import datetime
from sqlalchemy.sql import func
from collections import defaultdict

class DataLoader(object):
    '''
    DataLoader is used to load data from database to memory.
    (DB) -> DataLoader -> (Memory)
    '''

    DEFAULT_COMMISSION_CONF_FILE = '/shared/fof/config/sub_account.json'

    def __init__(self):
        '''
        Initialization
        '''
        self._logger = SurfingLogger.get_logger(type(self).__name__)
    def load(self, start_time=None, end_time=None, date=None, internal_data=None):
        '''
        Main function of DataLoader, to be called by DataEngine.
        :param start_time: start time for data loading
        :param end_time: end time for data loading
        :param date: date for daily update
        :return output: dict, {DataType -> data}
        '''
        output = {}
        output[DataType.INDUSTRY_INFO] = []
        output[DataType.INDUSTRY_PE] = []
        output[DataType.STOCK_FINANCIAL_INDICATOR] = []
        output[DataType.STOCK_DAILY_VALUE] = []
        output[DataType.STOCK_ANNUAL_NET_PROFIT] = []

        # Load industry division information
        industry_division = self._load_industry_info_data()
        if industry_division is not None:
            output[DataType.INDUSTRY_INFO] = industry_division
        self._logger.debug('Loaded industry division info')

        # Load stock pe ratio
        industry_pe = self._load_industry_pe()
        if industry_pe is not None:
            output[DataType.INDUSTRY_PE] = industry_pe
        self._logger.debug('Loaded industry pe ratio')


        if date is None:
            date = Timer.datetime()

        if date[-2:] == '01':
            # Load stock financial data
            stock_financial_data = self._load_stock_financial_indicator(date)
            if stock_financial_data is not None:
                output[DataType.STOCK_FINANCIAL_INDICATOR] = stock_financial_data
            self._logger.debug('Loaded stock financial data')

        else:
            # Load stock annual net profit
            stock_annual_net_profit = self._load_stock_latest_net_profit()
            if stock_annual_net_profit is not None:
                output[DataType.STOCK_ANNUAL_NET_PROFIT] = stock_annual_net_profit
            self._logger.debug('Loaded stock annual net profit')

        # Load stock daily value
        stock_daily_value = self._load_stock_valuation(date)
        if stock_daily_value is not None:
            output[DataType.STOCK_DAILY_VALUE] = stock_daily_value
        self._logger.debug('Loaded stock daily value')


        return output

    def _load_index_price_data_to_fund_indicators(self,start, end, order_book_id):
        '''
        获取某个指数一段时间内的价格
        :return: 某个指数一段时间内的价格, 若不存在为 None
        '''
        with QuantDatabaseConnector().managed_session() as derived_session:
            try:
                latest_record = derived_session.query(IndexPrice.ret,IndexPrice.date,IndexPrice.close).filter(
                    IndexPrice.date>= start, IndexPrice.date<=end, IndexPrice.order_book_id==order_book_id
                ).all()
                return latest_record
            except Exception as e:
                self._logger.error('获取某个指数一段时间内的价格失败 <err_msg>{}'.format(e))
        return None


    def _load_stock_valuation(self, date):
        '''
        获取某个时间点所有股票的市值
        :return: 某天所有股票的市值, 若不存在为 None
        '''
        with QuantDatabaseConnector().managed_session() as quant_session:
            try:
                latest_record = quant_session.query(StockValuation.order_book_id, StockValuation.date, StockValuation.market_cap_3).filter(
                    StockValuation.date == date
                ).all()
                return latest_record
            except Exception as e:
                self._logger.error('获取某个时间点所有股票的市值 <err_msg>{}'.format(e))
        return None


    def _load_industry_info_data(self):
        '''
        获取股票的行业分类
        :return: 全部股票的行业分类, 若不存在为 None
        '''
        with QuantDatabaseConnector().managed_session() as derived_session:
            try:
                latest_record = derived_session.query(StockInfo.industry_name,StockInfo.order_book_id).all()
                return latest_record
            except Exception as e:
                self._logger.error('获取获取股票的行业分类失败 <err_msg>{}'.format(e))
        return None

    def _load_industry_pe(self):
        '''
        获取行业的市盈率
        :return: 全部行业的名称和2015年来的历史市盈率, 若不存在为 None
        '''
        with DerivedDatabaseConnector().managed_session() as derived_session:
            try:
                latest_record = derived_session.query(IndustryPE.industry_name,IndustryPE.PE,IndustryPE.date,IndustryPE.P).all()
                #print(latest_record)
                return latest_record
            except Exception as e:
                self._logger.error('获取行业的市盈率 <err_msg>{}'.format(e))
        return None

    def _load_stock_latest_net_profit(self):
        '''
        获取股票最近年度净利润
        :return: 全部股票最近年度净利润, 若不存在为 None
        '''
        with DerivedDatabaseConnector().managed_session() as derived_session:
            try:
                latest_record = derived_session.query(StockLatestNetProfit.order_book_id, StockLatestNetProfit.update_date,StockLatestNetProfit.quarter, StockLatestNetProfit.net_profit).all()
                return latest_record
            except Exception as e:
                self._logger.error('获取股票最近年度净利润 <err_msg>{}'.format(e))
        return None

    def _load_stock_financial_indicator(self,date):
        '''
        获取股票最近财报
        :return: 全部股票最近财报, 若不存在为 None
        '''

        query_year = date[:4]
        pre_query_year = str(int(query_year)-2)
        with QuantDatabaseConnector().managed_session() as quant_session:
            try:
                latest_record = quant_session.query(StockFinancialIndicator.quarter, StockFinancialIndicator.order_book_id, StockFinancialIndicator.adjusted_net_profit,StockFinancialIndicator.announce_date).filter(
                    StockFinancialIndicator.quarter <= query_year, StockFinancialIndicator.quarter >=pre_query_year
                ).all()
                return latest_record
            except Exception as e:
                self._logger.error('获取股票最近财报 <err_msg>{}'.format(e))
        return None



if  __name__=='__main__':
    a = DataLoader()
    b = a.load(date='20190123')







# 接受一个要更新的日期参数，从数据库拉取行业行业分类信息，每支股票最新的净利润，历史PE数据，计算每个行业的PE，合并到历史PE数据并计算百分位后，上传更新原来的数据库

import pandas as pd
import time
import math

from fund.data_engine.data_processor import DpModule
from fund.util.utility import Utility
from fund.data_engine.common import DataType
from fund.db.models import IndustryPE, StockLatestNetProfit

# 流程
# 1. 每个月初更新profit table

# 2. 读取industry list(股票的行业分类信息)
# 3. 读取profit table（最近年度净利润）
# 4. 读取当天股票市值（财报公布之后才对股本数据进行调节，而不对报告期（例如，半年报为06-30）至公布日期之间的数据进行覆盖更新）
# 5. PE计算,更新PEQ
# 6. 写入industry_pe 和 stock_latest_pe数据库

class PeDailyModule(DpModule):
    '''
    计算杠杆率、爆仓价格距离和风险敞口
    '''

    def __init__(self):
        DpModule.__init__(self)
        self._necessary_keys = [
            DataType.INDUSTRY_INFO,
            DataType.INDUSTRY_PE,
            DataType.STOCK_FINANCIAL_INDICATOR,
            DataType.STOCK_DAILY_VALUE,
            DataType.STOCK_ANNUAL_NET_PROFIT
        ]

    def get_latest_annual_net_profit(self, df_stock_financial_indicator, stock):
        latest_financial_df = df_stock_financial_indicator[(df_stock_financial_indicator['order_book_id'] == stock)][
                              -5:]
        eps_df = latest_financial_df['adjusted_net_profit'].values
        quarter = latest_financial_df['quarter'].values
        if eps_df.shape[0] == 0:
            return 0, quarter[-1]
        else:
            for i in range(eps_df.shape[0] - 1, 0, -1):
                if quarter[i][-1] == '4' or quarter[i][-1] == '3' or quarter[i][-1] == '2':
                    eps_df[i] -= eps_df[i - 1]
            if eps_df.shape[0] < 5:
                # print('[info] NewStock')
                return eps_df.sum() / (eps_df.shape[0]) * 4, quarter[-1]
            else:
                return eps_df[1:].sum(), quarter[-1]

    def _run(self, data_in):
        '''
        计算行业估值
        :param data_in: 输入数据字典，需包括如下几个 keys
         - INDUSTRY_INFO: 行业划分信息
         - INDUSTRY_PE: 行业历史市盈率
         - STOCK_FINANCIAL_INDICATOR: 股票财报
         - STOCK_DAILY_VALUE: 股票每日市值
         - STOCK_ANNUAL_NET_PROFIT: 股票年度净利润
        '''
        # Output dict
        output = {}
        output[DataType.UPDATE_INDUSTRY_PE] = []
        # Get multiple data
        if data_in[DataType.STOCK_DAILY_VALUE] == []:
            self._logger.warn(f'No daily stock valuation data get!')
            print('No daily get!')
            return output

        industry_stock_dict = {}
        ind_all_df = pd.DataFrame(data_in[DataType.INDUSTRY_INFO])
        for _, i in ind_all_df.iterrows():
            if i['industry_name'] not in industry_stock_dict:
                industry_stock_dict[i['industry_name']] = [i['order_book_id']]
            else:
                industry_stock_dict[i['industry_name']].append(i['order_book_id'])
        df_industry_pe = pd.DataFrame(data_in[DataType.INDUSTRY_PE])
        if data_in[DataType.STOCK_FINANCIAL_INDICATOR] != []:
            df_stock_financial_indicator = pd.DataFrame(data_in[DataType.STOCK_FINANCIAL_INDICATOR])
            update_eps = True
            output[DataType.UPDATE_STOCK_ANNUAL_NET_PROFIT] = []
        else:
            update_eps = False
            df_stock_annual_net_profit = pd.DataFrame(data_in[DataType.STOCK_ANNUAL_NET_PROFIT])
        df_stock_valuation_daily = pd.DataFrame(data_in[DataType.STOCK_DAILY_VALUE])

        date = df_stock_valuation_daily['date'].values[0]

        # 如果计算了当天的pe，则删除历史数据重新计算
        df_industry_pe.drop(df_industry_pe[(df_industry_pe['date'] == date)].index, inplace=True)

        # 新建股票最新年度净值表
        if update_eps:
            df_stock_annual_net_profit = pd.DataFrame(
                {'order_book_id': [], 'update_date': [], 'quarter': [], 'net_profit': []})
        df_stock_annual_net_profit.set_index('order_book_id', inplace=True)

        # define a bref table to search
        df_stock_valuation_daily['new_index'] = df_stock_valuation_daily.loc[:,
                                                'date'] + df_stock_valuation_daily.loc[:, 'order_book_id']
        # stock_daily_price_df_short.apply(lambda x: x.date + x.order_book_id,axis=1)
        df_stock_valuation_daily.set_index('new_index', inplace=True)

        # 遍历上证指数更新日期
        for industry in industry_stock_dict:
            # 1.跳过无意义的行业
            if industry == "未知":
                continue
            e_industry = 0
            p_industry = 0
            pe_industry = 0
            num_stock = 0
            for stock in industry_stock_dict[industry]:
                # 2. 判断行业股票财报是否完全,并在当天有更新
                try:
                    if update_eps:
                        net_profit, quarter = self.get_latest_annual_net_profit(df_stock_financial_indicator, stock)
                        output[DataType.UPDATE_STOCK_ANNUAL_NET_PROFIT].append(
                            StockLatestNetProfit(order_book_id=str(stock), update_date=str(date),
                                                 quarter=str(quarter), net_profit=float(net_profit)))
                    else:
                        net_profit = float(df_stock_annual_net_profit.loc[stock, :]['net_profit'])
                    stock_close = float(df_stock_valuation_daily.loc[date + stock, :]['market_cap_3'])
                except Exception as e:
                    # print('No_information', industry, date + stock)
                    pass
                else:
                    if not math.isnan(stock_close):
                        e_industry += float(net_profit)
                        p_industry += stock_close
                        num_stock += 1
            pe_industry = p_industry / e_industry if e_industry != 0 else float('inf')
            # 计算p/e百分位
            peq = (pd.to_numeric(
                df_industry_pe.loc[(df_industry_pe.industry_name == industry) & (df_industry_pe.P != 0), 'PE']).append(
                pd.Series(pe_industry), ignore_index=True).rank(
                pct=True) * 100).values[-1]
            output[DataType.UPDATE_INDUSTRY_PE].append(
                IndustryPE(industry_name=industry, num_stock=num_stock, date=date, P=float(p_industry),
                           E=float(e_industry), PE=float(pe_industry), PEQ=float(peq)))
        return output

if __name__=='__main__':
    from python.data_engine.data_loader import DataLoader
    dl = DataLoader()
    data_in = dl.load(date='20190201')
    pe_cal = PeDailyModule()
    result = pe_cal._run(data_in)
    print(result)



import argparse
import sys

HELP_MSG = '''
fof <command> [<args>]

The most commonly used commands are:
    init_raw                 setup raw database
    init_basic               setup basic database
    init_derived             setup derived database
    init_view                setup view database
'''

class FundEntrance(object):

    def __init__(self):
        parser = argparse.ArgumentParser(description='Fund Command Tool', usage=HELP_MSG)
        parser.add_argument('command', help='Subcommand to run')
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print("Unrecognized command")
            parser.print_help()
            exit(1)
        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()

    def _get_yesterday(self):
        import datetime
        yesterday = datetime.datetime.today() - datetime.timedelta(days = 1)
        yesterday_str = datetime.datetime.strftime(yesterday, '%Y%m%d')
        print(f'yesterday: {yesterday_str}')
        return yesterday_str

    def init_raw(self):
        from .db.database import RawDatabaseConnector
        from .db.raw_models import Base

        print('Begin to create tables for database: raw')
        Base.metadata.create_all(bind=RawDatabaseConnector().get_engine())
        print('done')

    def init_basic(self):
        from .db.database import BasicDatabaseConnector
        from .db.basic_models import Base

        print('Begin to create tables for database: basic')
        Base.metadata.create_all(bind=BasicDatabaseConnector().get_engine())
        print('done')

    def init_derived(self):
        from .db.database import DerivedDatabaseConnector
        from .db.derived_models import Base

        print('Begin to create tables for database: derived')
        Base.metadata.create_all(bind=DerivedDatabaseConnector().get_engine())
        print('done')

    def init_view(self):
        from .db.database import ViewDatabaseConnector
        from .db.view_models import Base

        print('Begin to create tables for database: view')
        Base.metadata.create_all(bind=ViewDatabaseConnector().get_engine())
        print('done')

    def data_download(self):
        print('Start data download')
        from .downloader.raw_data_downloader import RawDataDownloader
        from .util.config import SurfingConfigurator
        rq_license = SurfingConfigurator().get_license_settings('rq')
        raw_data_downloader = RawDataDownloader(rq_license)
        yesterday = self._get_yesterday()
        raw_data_downloader.download(yesterday, yesterday)
        print('Done data download')

    def data_update(self):
        print('Start data update')

        print('Step 1. Start update basic')
        from .data_processor.basic.basic_data_processor import BasicDataProcessor
        basic_data_processor = BasicDataProcessor()
        yesterday = self._get_yesterday()
        basic_data_processor.process_all(yesterday, yesterday)
        print('Step 1. Done update basic')

        print('Done data update')
    

if __name__ == "__main__":
    entrance = FundEntrance()



import numpy as np
import pandas as pd
import datetime
from datetime import date
from pyfinance.ols import PandasRollingOLS
from fund.db.database import BasicDatabaseConnector,DerivedDatabaseConnector
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error, r2_score
from fund.data_api.basic import get_index_info, get_trading_day_list, get_index_price, get_fund_ret,get_fund_info,get_fund_nav
from fund.data_api.derived import get_fund_nav_adjusted_nv, get_index_benchmark_data,get_fund_list,get_fund_fee,get_fund_asset

class fund_indicators:
    
    # 优化日志
    # 计算前统一取数据
    # log return  <- normal return
    # 滚动定期回归算beta 定期 time_span， 不足time_span, beta nan
    # beta 计算 考虑无风险利率 基金每日ret -无风险利率/242（即每日无风险利率）
    # 日更 <- 一个季度算一次
    # 重现按照延报实现计算逻辑
    
    no_risk_ret = 0.03
    year = 242
    index_id = 'hs300'
    start_date  = date(2012,1,1)
    end_date = date(2020,3,30)
    
    long_time_span = 242 * 3 
    short_time_span = 242
    
    asset_dic = {
        '利率债' : 'national_debt',
        '美股大盘':'sp500rmb',
        '德国大盘':'dax30rmb',
        '信用债':'credit_debt',
        '房地产':'real_state',
        'A股大盘':'hs300',
        '原油': 'oil',
        '黄金': 'gold',
        '创业板':'gem',
        '日本大盘': 'n225rmb',
        'A股中盘':'csi500',
    }

    
    def __init__(self):
        
        self.data_prepare()
        self.build_history()
        self.upload_derived(self.result.dropna(), 'fund_indicator')
        
    def upload_derived(self, df, table_name):
        df.to_sql(table_name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
         
    def data_prepare(self):
        #self.fund_list = sorted(get_fund_list()['fund_id'].tolist())[:200]
        # asset exit
        fund_info = get_fund_info()[['fund_id','asset_type']]
        #TODO now only calculate funds which have tag 
        self.fund_list  = fund_info[fund_info['asset_type'] != 'null'].fund_id.tolist()
        #self.fund_list  = sorted(fund_info[fund_info['asset_type'] == 'null'].fund_id.tolist())[6000:]
        
        self.fund_nav = get_fund_nav_adjusted_nv(self.fund_list, self.start_date, self.end_date).drop_duplicates()
        self.fund_nav = self.fund_nav.pivot_table(index='datetime', columns='fund_id',values='adjusted_net_value')
        self.benchmark = get_index_benchmark_data([self.index_id], self.start_date, self.end_date)
        self.fund_fee = get_fund_fee() #基金费率
        self.fund_asset = get_fund_asset() # 基金资产类别
        asset_list = list(self.asset_dic.values())
        
        self.asset_data = get_index_benchmark_data(asset_list, self.start_date, self.end_date)
        self.asset_data = self.asset_data.pivot_table(index='datetime', columns='index_id',values='close')
        self.asset_data = np.log(self.asset_data) - np.log(self.asset_data.shift(1)) #各资产日收益 
        
        
        self.rm = self.benchmark[['close','datetime']].set_index('datetime')
        self.rm = np.log(self.rm) - np.log(self.rm.shift(1)) # 基准日收益log ret
        
        self.rb = self.no_risk_ret # 无风险利率 b -> bank
        self.rb_d = self.no_risk_ret / self.year # 无风险日利率
        self.rm_no_rb = self.rm - self.rb_d 
        self.rm_no_rb.columns = ['benchmark-risk_free'] # 基准日收益  - 无风险日利率
         
        self.rf = (np.log(self.fund_nav) - np.log(self.fund_nav.shift(1))) # 基金日收益log ret
        self.rf_no_rb = self.rf - self.rb_d # 基金日收益 - 日无风险利率 
        self.fund_info = get_fund_info()[['fund_id','asset_type','start_date','end_date']]
        
    def ployfit(self, x, df):
        # 返回多元回归系数 当前2
        df_i = df.reset_index().loc[x[0] : x[-1],]
        if set(df.head()['close'].isnull()) == {True}:
            return np.nan
        try:
            return np.polyfit(df_i['benchmark-risk_free'],df_i['close'],2)[0]
        except:
            return np.nan
    
    def rolling_beta(self, x, df, i ):
        # y在前 x在后
        # y =[2,4] ;x = [1,2]
        #np.polyfit(y,x,1)[0]  = 2
        df_i = df.reset_index().loc[x[0] : x[-1],].dropna()
        return np.polyfit(df_i[i],df_i['benchmark-risk_free'],1)[0]
        
    def get_beta(self, t_range, fund_list):
        # beta 贝塔 
        # 表示基金的收益相对与基准对变化， 绝对值越大， 相对于基准波动越大， 体现为风险越大
        # rf_no_rb = 基金日收益 - 无风险日收益  
        # rm_no_rb = 基准日收益 - 无风险日收益
        # v1 对 v2 做回归 beta是一次项系数   rf_no_rb = beta * rm_no_rb + C （滚动计算）
        # 回归的时间长度取 time_span
        res = []
        date_list = self.rf_no_rb.index
        for i in fund_list:
            try:
                df = self.rf_no_rb[[i]].join(self.rm_no_rb).dropna()
                x = df['benchmark-risk_free']
                y = df[i]
                model = PandasRollingOLS(y=y, x=x, window=t_range)
                res_i = model.beta
                res_i.columns = [i]
                res.append(res_i)
            except:
                continue
        return pd.concat(res, axis =1, sort=True)
        
    def get_treynor(self):
        # treynor 特雷诺
        # 表示评价时间范围内 基金单位风险对无风险收益
        # rf = 基金一段时间平均收益 - 无风险收益 （根据参考时间做年化 这里日收益* time_span） （滚动计算）
        # beta 参考上面
        # treynor = rf / beta
        
        tr_up = self.rf_avg
        tr_up = tr_up.loc[self.beta.index,].copy() * self.time_span -  self.rb_d * self.time_span
        self.treynor = tr_up / self.beta
        self.treynor = self.treynor.loc[self.beta.index[0]:,]
        
    def get_alpha(self, fund_list, beta):
        # alpha 阿尔法
        # 表示超越市场对收益
        # rf_avg = 基金评价期间平均日收益  （滚动计算）
        # rb_d = 无风险日利率 
        # rm_avg = 基准评价期间平均日收益  （滚动计算）
        # alpha = (rf_avg - rb_d - beta*(rm_avg - rb_d) ) * year
        c = [_ for _ in self.rf_avg.columns if _ in fund_list]
        v1 = (self.rm_avg - self.rb_d).loc[beta.index,].copy()
        v2 = beta.multiply(v1.values, axis='close').copy()
        alpha = ((self.rf_avg[c] - self.rb_d - v2) * self.year).copy()
        alpha = alpha.loc[beta.index[0]:,].copy()
        return alpha
    
    def get_mdd(self):
        # mdd 最大回撤
        # pandas :  1 - (df / df.cummax()).min()
        mdd_fuc = lambda x: 1 - (x /np.maximum.accumulate(x)).min(axis=0)
        self.mdd = self.fund_nav.rolling(self.time_span).apply(mdd_fuc, raw=True)
        self.mdd = self.mdd.loc[self.beta.index[0]:,]
    
    def get_downside_risk(self):
        # downside risk 下行风险
        # sqrt(year) / T * sum( min(0, rf - rb_d))
        df = self.rf - self.rb_d
        df = (np.sign(df)*-1+1 )/2 * df * -1
        self.down_risk = df.rolling(self.time_span).sum() * np.sqrt(self.time_span / self.year) / self.year
        self.down_risk = self.down_risk.loc[self.beta.index[0]:,]
        
    def get_return_over_period(self):
        self.ret_over_period = np.log(self.fund_nav) - np.log(self.fund_nav.shift(self.time_span))
        self.ret_over_period = self.ret_over_period.loc[self.beta.index[0]:,]
    
    def get_annulized_average_daily_return(self):
        self.annual_avg_daily_ret = (self.rf.rolling(self.time_span).sum()*self.year/self.time_span).loc[self.beta.index,]
        self.annual_avg_daily_ret = self.annual_avg_daily_ret.loc[self.beta.index[0]:,]
    
    def get_volatility(self):
        self.vol = self.rf.rolling(self.time_span).std().loc[self.beta.index,]
        self.vol = self.vol.loc[self.beta.index[0]:,]
    
    def get_m_square(self):
        # m_square  m2测度
        # 风险调整指标，与sharpe完全线性关系，在排序上与sharpe等价
        # 含义更直观，同风险下相对与市场组合的超额收益率
        # rf_avg = 基金在评价期的平均收益率
        # vol = 波动率 收益率标准差
        # rm_avg = 评价期上市场指数平均收益率
        # vol_m = 市场指数波动率
        # rb_d = 无风险日利率
        # m2 = (vol_m / vol * (rf_avg - rb_d) + rb_d) - rm_avg
        v1 = (self.rf_avg - self.rb_d).loc[self.beta.index,]
        v2 = self.vol_m.loc[self.beta.index,]
        v3 = v1.multiply(v2.values, axis='close')
        v4 = v3 / self.vol
        self.m_square = (v4 + self.rb_d).sub(self.rm_avg.loc[self.beta.index,].values, axis='close') 
        self.m_square = self.m_square.loc[self.beta.index[0]:,]
    
    def get_time_return(self):
        # time_return 择时收益
        # 择时收益是一种业绩归因指标，衡量基金在评价期上通过择时获得的超额收益
        # beta_i 基金在自区间beta
        # beta_0 基金在评价期上的beta
        # rm_i 基准在区间收益率
        # rb_i 无风险资产在自区间收益率
        # 区间 当前取 3M ->60
        # time_ret = sum[(beta_i - beta_0) * (rm_i - rb_i)]
        self.day_range = 60
        rb_i = self.no_risk_ret / self.year * self.day_range
        self.beta_i = self.get_beta(self.day_range)
        self.rm_i = self.rm.rolling(self.day_range).sum()
        self.rm_i = self.beta_i.join(self.rm_i).fillna(method = 'ffill')[['close']].copy()
        date_list = self.beta_i.index.tolist()
        i_num = int(self.time_span / self.day_range) - 1

        res = []
        for d in self.beta.index:
            idx = date_list.index(d)
            time_ret = 0
            for i in range(i_num):
                idx_i = idx  - (1+i) * self.day_range
                d_i = date_list[idx_i]
                v1 = self.beta_i.loc[d_i,] - self.beta.loc[d,]
                v2 = self.rm_i.loc[d_i,] - rb_i
                time_ret_i = v1 * v2.values
                time_ret += time_ret_i
            res.append(time_ret.copy())
        self.time_ret = pd.DataFrame(res)
        self.time_ret.index = self.beta.index
        
    def get_value_at_risk(self):
        # value at risk 资产在险值
        # 在评价期间内，基金95%的日跌幅不会大于该数值
        # var = -min{0, r(t) 大到小排序 95分位数}
        self.var = self.rf.rolling(self.time_span).quantile(0.05)
        self.var[self.var>0] = 0
        self.var = -1 * self.var
        self.var = self.var.loc[self.beta.index[0]:,]
    
    def get_r_square(self):
        # r2 决定系数r方
        # 衡量基金收益波动的方差中，市场决定比例， 简单理解市场风险占总风险权重
        res = []
        date_list = self.rf_no_rb.index
        for i in self.fund_list:
            df = self.rf_no_rb[[i]].join(self.rm_no_rb).fillna(method = 'ffill')[1:]
            x = df['benchmark-risk_free']
            y = df[i]
            model = PandasRollingOLS(y=y, x=x, window=self.time_span)
            res.append(model.rsq)
        df = pd.DataFrame(res).T
        df.columns = self.fund_list
        df.index   = self.rf_no_rb.index[-df.shape[0]:]    
        self.r_square = df.loc[self.beta.index[0]:,]
        
    def get_sharpe_ratio(self):
        # sharpe 夏普
        # 简单年化 ret * year , vol * sqrt( year )
        sr_ret = self.rf_no_rb.rolling(self.time_span).mean() * self.year
        sr_vol = self.rf.rolling(self.time_span).std() * np.sqrt(self.year)
        self.sharpe = (sr_ret/ sr_vol).loc[self.beta.index[0]:,]
    
    def get_treynor_mazuy_coefficient(self):
        # tm 系数描述基金在评价期上的择时收益
        # 以 rf_no_rb 为因变量对  rm_no_rb 及其平方进行回归的二次项系数
        # 以df的index 序数做 rolling apply, 输入多个列的， 可以此法实现pyfinance.PandasRollingOLS
        res = []
        date_list = self.rf_no_rb.index
        for i in self.fund_list:
            df = self.rf_no_rb[[i]].join(self.rm_no_rb).fillna(method = 'ffill')[1:]
            x = df['benchmark-risk_free']
            y = df[i]
            y.name = 'close'
            df = pd.DataFrame([x,y]).T
            df['idx'] = range(df.shape[0])
            res_i = df[['idx']].rolling(self.time_span).apply(lambda x : self.ployfit(x, df), raw= True)
            res_i.columns = [i]
            res.append(res_i)
        self.tm_coef = pd.concat(res, axis = 1).loc[self.beta.index[0]:,]
        
    def get_fee_rate(self):
        self.fee_rate = self.fund_fee[self.fund_fee['fund_id'].isin(self.fund_list)]
        self.fee_rate.loc[:,'fee_rate'] = ((self.fee_rate['manage_fee'] + self.fee_rate['trustee_fee'])/100).copy()
        self.fee_rate = self.fee_rate[['fund_id', 'fee_rate']]
        
    def get_track_error(self, t_range, fund_list, beta):
        df_asset = self.fund_asset[self.fund_asset['fund_id'].isin(fund_list)]
        con_cal = df_asset['asset_type'] != 'null'
        con_pas = df_asset['asset_type'] == 'null'
        fl_pas = df_asset[con_pas]['fund_id']
        res = []
        c = []
        for idx,f in df_asset[con_cal].iterrows(): 
            try:
                f_i = f.fund_id
                a_t = f.asset_type
                a_i = self.asset_dic[a_t]
                ret_b = self.asset_data[[a_i]]
                ret_f = self.rf[[f_i]]
                df_tmp = ret_f.join(ret_b).fillna(0)
                df_tmp = (df_tmp[f_i] - df_tmp[a_i]).rolling(t_range).std()*np.sqrt(242).copy()
                df_tmp.columns = [f_i]
                res.append(df_tmp.copy())
                c.append(f_i)
            except:
                continue
        
        track_err = pd.concat(res, axis = 1, sort=True)
        track_err.columns = c
        for i in fl_pas:
            track_err[i] =  np.nan
        track_err = track_err.loc[beta.index[0]:,]
        return track_err

    def build_history(self):
        # ['A股中盘','A股大盘','创业板'] 3年 历史数据 注重alpha 
        # 其他 用1年  成立较晚 注重track_err 
        fund_info = self.fund_info
        fund_info = fund_info[fund_info['asset_type'] != 'null']
        type_lterm = ['A股中盘','A股大盘','创业板']
        asset_list = list(self.asset_dic.keys()) 
        type_sterm = [_ for _ in asset_list if not _ in type_lterm]
        fl_lterm  = fund_info[fund_info['asset_type'].isin(type_lterm)].fund_id.tolist()
        fl_sterm  = fund_info[fund_info['asset_type'].isin(type_sterm)].fund_id.tolist()
        self.get_fee_rate()
        result_df = []
        for fl, time_span in zip([fl_lterm, fl_sterm],  [ self.long_time_span,  self.short_time_span]):
            self.rm_avg = self.rm.rolling(time_span).mean()# 基准评价期间平均日收益
            self.vol_m = self.rm.rolling(time_span).std() # 基准评价期间波动率
            self.rf_avg = self.rf.fillna(0).rolling(time_span).mean() # 基金评价期间平均日收益
            
            beta = self.get_beta(time_span, fl)
            alpha = self.get_alpha(fl, beta)
            track_err = self.get_track_error(time_span, fl, beta)
            date_last = track_err.index[-1]
            fact_l = ['beta','alpha','track_err']
            data_l = [beta, alpha, track_err]
            res = []
            for fun_i in fl:
                for f, d in zip(fact_l, data_l):
                    if fun_i not in d.columns:
                        continue
                    d = d.loc[:date_last][[fun_i]].stack().reset_index(level=[0,1]).copy()
                    d.columns = ['datetime','fund_id','value']
                    d['name'] = f
                    res.append(d.copy())
            result = pd.concat(res).pivot_table(index=['datetime','fund_id'], columns='name',values='value')
            result = result.join(self.fee_rate.set_index('fund_id'))   
            result = result.reset_index(level=[0,1])
            result['timespan'] = time_span / self.year
            result = result.replace(np.inf,np.nan)
            result = result.replace(-np.inf,np.nan)
            result_df.append(result)
        self.result = pd.concat(result_df).dropna()
        
f_i = fund_indicators()



import os
from ..util.config import SurfingConfigurator
from tlclient.linker.logger import Logger

class SurfingLogger(Logger):

    @classmethod
    def _get_log_path_from_env(cls):
        log_path = SurfingConfigurator().get_sys_settings().get_log_dir()
        if not os.path.exists(log_path):
            os.makedirs(log_path)
        
        return log_path



import datetime
import copy
import json
import numpy as np
import pandas as pd
import os

from .asset_helper import SAAHelper, TAAHelper
from .trader import AssetTrader
from .report import BacktestInfo
from ...data.manager.manager_fund import FundDataManager
from ...data.struct import AssetWeight, AssetPrice, AssetPosition, AssetValue


class FundBacktestEngine:

    def __init__(self, data_manager: FundDataManager):
        self._dm = data_manager
        self._saa_helper = SAAHelper() # 战略目标
        self._taa_helper = TAAHelper()
        self._report_helper = BacktestInfo()
        self._asset_trader = AssetTrader()

    def init(self):
        self._dm.init()
        self._saa_helper.init()
        self._taa_helper.init()
        self._report_helper.init()
        self._asset_trader.init()

    def run(self, saa: AssetWeight, start_date: datetime.date=None, end_date: datetime.date=None):
        # init position
        self.cur_asset_position = AssetPosition(cash=1e8)
        cur_asset_mv = AssetValue(prices=AssetPrice(cash=1), positions=self.cur_asset_position)
        print(cur_asset_mv)
        self._pending_trades = []

        # init days
        start_date = start_date or self._dm.start_date
        end_date = end_date or self._dm.end_date

        # setup helpers
        self._saa_helper.setup(saa)
        self._taa_helper.setup(saa)
        self._report_helper.setup(saa)
        self._asset_trader.setup(saa)

        # loop trading days
        _dts = self._dm.get_trading_days()
        dts = _dts[(_dts.datetime >= start_date) & (_dts.datetime <= end_date)].datetime # df with datetime.date
        for t in dts:
            self._run_on(t)
        
    def _run_on(self, dt):
        # current asset price
        cur_asset_price = self._dm.get_index_price_data(dt)
        self._pending_trades = self._asset_trader.finalize_trade(dt, self._pending_trades, cur_asset_price, self.cur_asset_position)
        target_asset_allocation = self.calc_asset_allocation(dt, cur_asset_price)

        virtual_position, trade_list = self._asset_trader.calc_trade(dt, self.cur_asset_position, cur_asset_price, target_asset_allocation)
        if trade_list and len(trade_list) > 0:
            self.book_trades(trade_list)
        self._report_helper.update(dt, self.cur_asset_position, cur_asset_price, self._pending_trades)

    def calc_asset_allocation(self, dt, cur_asset_price: AssetPrice):
        cur_saa = self._saa_helper.on_price(dt, cur_asset_price)
        asset_pct = self._dm.get_index_pcts(dt)
        cur_taa = self._taa_helper.on_price(dt, cur_asset_price, cur_saa, asset_pct)
        return cur_taa

    def book_trades(self, trade_list: list):
        self._pending_trades += trade_list

    def report(self):
        result = self._report_helper.result()# backtest result
        index_df = self._dm.get_index_price()
        self._report_helper.backtest_plot(index_df)# backtest plot

if __name__ == '__main__':
    m = FundDataManager('20100101', '20200101')
    saa = AssetWeight(
        hs300=15/100,
        csi500=5/100,
        gem=3/100,
        sp500rmb=7/100,
        national_debt=60/100,
        gold=10/100,
        cash=5/100
    )
    b = FundBacktestEngine(data_manager=m)
    b.init()
    b.run(saa=saa, start_date=datetime.date(2010,1,1), end_date=datetime.date(2020,1,1))
    #b.report()


import pandas as pd
import datetime
from ...db.database import BasicDatabaseConnector
from ...data_api.raw_new import RawDataApi
from ...data_api.basic_new import BasicDataApi
from ...db.basic_models import *

class BasicDataProcessor(object):

    def __init__(self):
        self._index_info_df = BasicDataApi().get_index_info()
        self._fund_info_df = BasicDataApi().get_fund_info()
        # Get the mapping between order_book_id and fund_id (which is defined in this system)
        self._fund_id_mapping = BasicDataApi().get_fund_id_mapping()

    def _upload_basic(self, df, table_name, if_exists='append'):
        print(table_name)
        # print(df)
        df.to_sql(table_name, BasicDatabaseConnector().get_engine(), index=False, if_exists=if_exists)

    def _get_fund_id_from_order_book_id(self, order_book_id, date):
        if order_book_id in self._fund_id_mapping:
            for fund_info in self._fund_id_mapping[order_book_id]:
                if (date >= fund_info['start_date'] and 
                        (fund_info['end_date'] == '00000000' or date <= fund_info['end_date'])):
                    return fund_info['fund_id']
        return None

    def _status_mapping(self, x):
        if x == 'Open':
            return 0
        elif x == 'Suspended':
            return 1
        elif x == 'Limited':
            return 2
        elif x == 'Close':
            return 3
        else:
            return None

    def _find_tag(self, symbol, wind_class_II):
        if '沪深300' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return 'A股大盘'
        elif '中证500' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return 'A股中盘'
        elif '标普500' in symbol:
            return '美股大盘'
        elif '创业板' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return '创业板'
        elif '德国' in symbol:
            return '德国大盘'
        elif '日本' in symbol or '日经' in symbol:
            return '日本大盘'
        elif (('国债' in symbol or '利率' in symbol or '金融债' in symbol)
              and wind_class_II in ['短期纯债型基金', '中长期纯债型基金', '被动指数型债券基金']):
            return '利率债'
        elif (('信用' in symbol or '企债' in symbol or '企业债' in symbol)
              and wind_class_II in ['短期纯债型基金', '中长期纯债型基金', '被动指数型债券基金']):
            return '信用债'
        elif '黄金' in symbol:
            return '黄金'
        elif '原油' in symbol or '石油' in symbol or '油气' in symbol:
            return '原油'
        elif ('地产' in symbol or '金融' in symbol) and ('美国'not in symbol):
            return '房地产'
        else:
            return 'null'

    def fund_info(self):
        # Update manually
        # Not verified
        track_index_df = pd.read_csv('./data/fund_track_index.csv',index_col=0 )
        wind_fund_info = RawDataApi().get_wind_fund_info()
        fund_fee = RawDataApi().get_fund_fee()
        wind_fund_info['order_book_id'] = [_.split('!')[0].split('.')[0] for _ in wind_fund_info['wind_id'] ]
        res = []
        for i in wind_fund_info['wind_id']:
            if not '!' in i:
                res.append(0)
            else:
                res_i = int(i.split('!')[1].split('.')[0])
                res.append(res_i)
        wind_fund_info['transition'] = res
        wind_fund_info['fund_id'] = [o + '!' + str(t) for o, t in zip(wind_fund_info['order_book_id'], wind_fund_info['transition'])]
        wind_fund_info = wind_fund_info.set_index('fund_id')
        fund_fee = fund_fee.drop(['id','desc_name'], axis = 1).set_index('fund_id')
        wind_fund_info = wind_fund_info.join(fund_fee)
        wind_fund_info['fund_id'] = wind_fund_info.index
        wind_fund_info['asset_type'] = [self._find_tag(symbol, wind_class_II) for symbol, wind_class_II in zip(wind_fund_info['desc_name'],wind_fund_info['wind_class_2'])]
        wind_fund_info['update_time'] = datetime.datetime.now()
        wind_fund_info = wind_fund_info.drop(['id'], axis = 1)
        wind_fund_info = wind_fund_info[[i.split('.')[1] == 'OF' for i in wind_fund_info['wind_id']]]
        wind_fund_info = wind_fund_info.set_index('wind_id')
        dic = {k:v for k,v in zip(self._index_info_df['desc_name'], self._index_info_df['index_id'])}
        wind_fund_info = wind_fund_info.join(track_index_df[['track_index']])
        wind_fund_info['wind_id'] = wind_fund_info.index
        wind_fund_info['track_index'] = wind_fund_info['track_index'].map(lambda x: dic.get(x,'none'))
        self._upload_basic(wind_fund_info, FundInfo.__table__.name)

    def index_info(self):
        # Update manually
        # Not verified
        df = pd.read_csv('./data/index_info.csv')
        self._upload_basic(df, IndexInfo.__table__.name)

    def fund_nav(self, start_date, end_date):
        df = RawDataApi().get_fund_nav(start_date, end_date)
        df['fund_id'] = df.apply(
            lambda x: self._get_fund_id_from_order_book_id(x['order_book_id'], x['datetime']), axis=1)
        df['subscribe_status'] = df['subscribe_status'].map(self._status_mapping)
        df['redeem_status'] = df['redeem_status'].map(self._status_mapping)
        df = df[df['fund_id'].notna()]
        self._upload_basic(df, FundNav.__table__.name)

    def index_price(self, start_date, end_date):
        cm_index_price = RawDataApi().get_raw_cm_index_price_df(start_date, end_date)
        cxindex_index_price = RawDataApi().get_cxindex_index_price_df(start_date, end_date)
        yahoo_index_price = RawDataApi().get_yahoo_index_price_df(start_date, end_date)
        rq_index_price = RawDataApi().get_rq_index_price_df(start_date, end_date)

        df_list = []
        index_list = ['sp500', 'dax30', 'n225']
        cm_index_list = ['sp500rmb', 'dax30rmb', 'n225rmb']
        for i, c in zip(index_list, cm_index_list):
            cm_index = yahoo_index_price.copy()
            df = cm_index[cm_index['index_id'] == i]
            df = cm_index_price.set_index('datetime').join(df.set_index('datetime'))
            df = df.fillna(method='ffill')
            df['close'] = df['close'] * df['usd_central_parity_rate']
            df = df[['close']]
            df['datetime'] = df.index
            df['ret'] = df['close'] / df['close'].shift(1) - 1
            df['open'] = float('Nan')
            df['high'] = float('Nan')
            df['low'] = float('Nan')
            df['volume'] = float('Nan')
            df['total_turnover'] = float('Nan')
            df['index_id'] = c
            df_list.append(df.copy())

        index_dic = {k: v for k, v in zip(
            self._index_info_df['order_book_id'], self._index_info_df['index_id']) if k != 'not_available'}
        rq_index_price['index_id'] = rq_index_price['order_book_id'].map(
            lambda x: index_dic[x])
        res = []
        for index_i in list(set(rq_index_price['order_book_id'].tolist())):
            dftmp = rq_index_price[rq_index_price['order_book_id'] == index_i].copy(
            )
            dftmp['ret'] = dftmp['close'] / dftmp['close'].shift(1) - 1
            res.append(dftmp)
        rq_index_price = pd.concat(res)
        df_list.append(yahoo_index_price)
        df_list.append(cxindex_index_price)
        df_list.append(rq_index_price)

        df = pd.concat(df_list).drop(['order_book_id'], axis=1)
        df['datetime'] = pd.to_datetime(df['datetime'])
        df = df.drop(['id'], axis=1)
        self._upload_basic(df, IndexPrice.__table__.name)

    def fund_rating_latest(self):
        fund_rating_df = RawDataApi().get_fund_rating()
        fund_info = self._fund_info_df[['fund_id', 'order_book_id', 'start_date', 'end_date']]
        score_name = ['zs', 'sh3', 'sh5', 'jajx']
        score_dic = {}
        for s in score_name:
            score_df = fund_rating_df[['order_book_id', 'datetime', s]]
            datelist = sorted(list(set(score_df.dropna().datetime.tolist())))
            score_dic[s] = datelist[-1]
        res = []
        for s, d in score_dic.items():
            df = fund_rating_df[['order_book_id', 'datetime', s]]
            df = df[df['datetime'] == d]
            con1 = fund_info['start_date'] <= d
            con2 = fund_info['end_date'] >= d
            fund_info_i = fund_info[con1 & con2]
            dic = {row['order_book_id']: row['fund_id']
                   for index, row in fund_info_i.iterrows()}
            df['fund_id'] = df['order_book_id'].map(lambda x: dic[x])
            df = df[['fund_id', s]].copy().set_index('fund_id')
            res.append(df)
        df = pd.concat(res, axis=1, sort=False)
        df['fund_id'] = df.index
        df['update_time'] = datetime.date.today()
        self._upload_basic(df, FundRatingLatest.__table__.name, 'replace')

    def stock_price(self, start_date, end_date):
        stock_price_df = RawDataApi().get_rq_stock_price(start_date, end_date)
        stock_post_price_df = RawDataApi().get_rq_stock_post_price(start_date, end_date)

        stock_post_price_df['adj_close'] = stock_post_price_df['close']
        stock_post_price_df = stock_post_price_df.filter(items=['adj_close', 'datetime', 'order_book_id'], 
            axis='columns')

        stock_price_merge_df = stock_price_df.merge(stock_post_price_df, how='left', on=['datetime', 'order_book_id'])
        stock_price_merge_df['post_adj_factor'] = stock_price_merge_df['adj_close'] / stock_price_merge_df['close']
        stock_price_merge_df['ret'] = stock_price_merge_df['adj_close'] / stock_price_merge_df['adj_close'].shift(1) - 1
        stock_price_merge_df = stock_price_merge_df.rename(columns={'order_book_id': 'stock_id'}).drop(
            columns=['id'])

        self._upload_basic(stock_price_merge_df, StockPrice.__table__.name)

    def fund_ret(self, start_date, end_date):
        column_rename_dict = {
            'sharp_ratio': 'sharpe_ratio',
            'last_week_return': 'w1_ret',
            'last_month_return': 'm1_ret',
            'last_three_month_return': 'm3_ret',
            'last_six_month_return': 'm6_ret',
            'last_twelve_month_return': 'y1_ret',
            'max_drop_down': 'mdd',
            'annualized_returns': 'annual_ret',
            'average_size': 'avg_size',
            'information_ratio': 'info_ratio',
            'to_date_return':'to_date_ret'
        }
        column_drop_list = ['id', 'order_book_id', 'year_to_date_return', 'annualized_risk']

        df = RawDataApi().get_rq_fund_indicator(start_date, end_date)
        df['fund_id'] = df.apply(
            lambda x: self._get_fund_id_from_order_book_id(x['order_book_id'], x['datetime']), axis=1)
        df = df.rename(columns=column_rename_dict).drop(columns=column_drop_list)

        self._upload_basic(df, FundRet.__table__.name)

    def stock_info(self):
        df = RawDataApi().get_stock_info()
        self._upload_basic(df, StockInfo.__table__.name, 'replace')

    def fund_size(self):
        df = RawDataApi().get_fund_size()
        df['fund_id'] = df.apply(
            lambda x: self._get_fund_id_from_order_book_id(x['order_book_id'], x['update_time']), axis=1)
        df = df.drop(['order_book_id'], axis = 1)
        df = df[df['fund_id'].notnull()]
        self._upload_basic(df, FundSize.__table__.name, 'replace')

    def trading_day_list(self, start_date, end_date):
        df = RawDataApi().get_trading_day_list(start_date, end_date)
        self._upload_basic(df, TradingDayList.__table__.name)
    
    def process_all(self, start_date, end_date):
        self.fund_nav(start_date, end_date)
        self.fund_ret(start_date, end_date)
        self.index_price(start_date, end_date)
        self.stock_price(start_date, end_date)
        self.trading_day_list(start_date, end_date)
        self.stock_info()
        self.fund_size()
        self.fund_rating_latest()



from .message_receiver import WechatMessageReceiver
from .message_parser import WechatMessageParser
from .message_sender import WechatMessageSender
import time
import json
import urllib.parse
import sys
from .config import Configurator

class WechatMessageProcessor(object):

    def __init__(self):
        self.message_receiver = WechatMessageReceiver()
        self.message_parser = WechatMessageParser()
        self.message_sender = WechatMessageSender()

    def run(self):
        while True:
            time.sleep(0.01)

            message = self.message_receiver.get_message()
            if not message:
                continue
            
            result = self.message_parser.parse(message["msg_from"], message["msg_to"], message["msg_content"])

            # we can handle scenario 1 only
            if result['scenario_id'] == 1:
                target_url = 'http://52.82.113.3:12300/FundInfo?id='
                content = target_url + urllib.parse.quote_plus(result['param']['fund_id'])
            else:
                content = json.dumps(result)

            self.message_sender.send(message["msg_to"], content)

            #print(f'msg: ({message})\nresult: ({json.dumps(result)})\n')

if __name__ == "__main__":
    if len(sys.argv) > 1:
        config_path = sys.argv[1]
        print(f'config path is set: {config_path}')
        Configurator(config_path).config_path = config_path
    wmp = WechatMessageProcessor()
    wmp.run()


from sqlalchemy import CHAR, Column, Integer
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import DOUBLE, DATE, TINYINT

Base = declarative_base()

class IndexVolatility(Base):
    '''指数波动率'''

    __tablename__ = 'index_volatility'
    id = Column(Integer, primary_key=True)

    index_id = Column(CHAR(20)) #指数ID
    datetime = Column(DATE) # 日期
    w1_vol = Column(DOUBLE(asdecimal=False)) # 近一周波动率
    m1_vol = Column(DOUBLE(asdecimal=False)) # 近一月波动率
    m3_vol = Column(DOUBLE(asdecimal=False)) # 近三月波动率
    m6_vol = Column(DOUBLE(asdecimal=False)) # 近半年波动率
    y1_vol = Column(DOUBLE(asdecimal=False)) # 近一年波动率
    y3_vol = Column(DOUBLE(asdecimal=False)) # 近三年波动率
    y5_vol = Column(DOUBLE(asdecimal=False)) # 近五年波动率
    y10_vol = Column(DOUBLE(asdecimal=False)) # 近十年波动率
    this_y_vol = Column(DOUBLE(asdecimal=False)) # 今年以来波动率
    cumulative_vol = Column(DOUBLE(asdecimal=False)) # 成立至今波动率


class IndexReturn(Base):
    '''指数波动率'''

    __tablename__ = 'index_return'
    id = Column(Integer, primary_key=True)

    index_id = Column(CHAR(20)) #指数ID
    datetime = Column(DATE) # 日期
    w1_ret = Column(DOUBLE(asdecimal=False)) # 近一周收益率
    m1_ret = Column(DOUBLE(asdecimal=False)) # 近一月收益率
    m3_ret = Column(DOUBLE(asdecimal=False)) # 近三月收益率
    m6_ret = Column(DOUBLE(asdecimal=False)) # 近半年收益率
    y1_ret = Column(DOUBLE(asdecimal=False)) # 近一年收益率
    y3_ret = Column(DOUBLE(asdecimal=False)) # 近三年收益率
    y5_ret = Column(DOUBLE(asdecimal=False)) # 近五年收益率
    y10_ret = Column(DOUBLE(asdecimal=False)) # 近十年收益率
    this_y_ret = Column(DOUBLE(asdecimal=False)) # 今年以来收益率
    cumulative_ret = Column(DOUBLE(asdecimal=False)) # 成立至今收益率


class FundAlpha(Base):
    '''基金超额收益'''

    __tablename__ = 'fund_alpha'
    id = Column(Integer, primary_key=True)

    track_err	= Column(DOUBLE(asdecimal=False)) # 跟踪误差
    this_y_alpha	= Column(DOUBLE(asdecimal=False)) # 今年以来超额收益
    cumulative_alpha	= Column(DOUBLE(asdecimal=False)) # 成立以来超额收益
    w1_alpha	= Column(DOUBLE(asdecimal=False)) # 近一周收益率
    m1_alpha    = Column(DOUBLE(asdecimal=False)) # 近一月超额收益
    m3_alpha	= Column(DOUBLE(asdecimal=False)) # 近三月超额收益
    m6_alpha	= Column(DOUBLE(asdecimal=False)) # 近半年超额收益
    y1_alpha	= Column(DOUBLE(asdecimal=False)) # 近一年超额收益
    y3_alpha	= Column(DOUBLE(asdecimal=False)) # 近三年超额收益
    y5_alpha	= Column(DOUBLE(asdecimal=False)) # 近五年超额收益
    y10_alpha	= Column(DOUBLE(asdecimal=False)) # 近十年超额收益
    fund_id	= Column(CHAR(20)) #基金ID
    datetime= Column(DATE) # 日期


class IndexValuation(Base):
    '''基金超额收益'''

    __tablename__ = 'index_valuation'
    id = Column(Integer, primary_key=True)
    
    index_id = Column(CHAR(20)) #指数ID
    pb_mrq = Column(DOUBLE(asdecimal=False)) # 市净率-MRQ
    pe_ttm = Column(DOUBLE(asdecimal=False)) # 市盈率-MMT
    peg_ttm = Column(DOUBLE(asdecimal=False)) # PEG-MMT
    roe_ttm = Column(DOUBLE(asdecimal=False)) # 净资产收益率-MMT
    dy_ttm = Column(DOUBLE(asdecimal=False)) # 股息率-MMT
    pe_pct = Column(DOUBLE(asdecimal=False)) # PE百分位
    pb_pct = Column(DOUBLE(asdecimal=False)) # PB百分位
    val_score = Column(DOUBLE(asdecimal=False)) # 估值评分
    datetime = Column(DATE) # 日期


class FundIndicator(Base):
    '''基金评价指标'''

    __tablename__ = 'fund_indicator'
    id = Column(Integer, primary_key=True)

    fund_id = Column(CHAR(10)) #基金ID
    datetime =  Column(DATE) # 日期

    beta = Column(DOUBLE(asdecimal=False)) #风险指数
    alpha = Column(DOUBLE(asdecimal=False)) #投资回报
    track_err = Column(DOUBLE(asdecimal=False)) #跟踪误差
    timespan = Column(DOUBLE(asdecimal=False)) #历史数据跨度(年)
    fee_rate = Column(DOUBLE(asdecimal=False)) #费率
    # treynor = Column(DOUBLE(asdecimal=False)) #特雷诺比率 
    # mdd = Column(DOUBLE(asdecimal=False)) #净值最大回撤
    # down_risk = Column(DOUBLE(asdecimal=False)) #下行风险
    # ret_over_period = Column(DOUBLE(asdecimal=False)) #区间收益率
    # annual_avg_daily_ret = Column(DOUBLE(asdecimal=False)) #年化日均收益
    # vol = Column(DOUBLE(asdecimal=False)) #波动率
    # m_square = Column(DOUBLE(asdecimal=False)) #M平方测度 风险调整收益指标
    # time_ret = Column(DOUBLE(asdecimal=False)) #择时收益
    # var = Column(DOUBLE(asdecimal=False)) #资产在险值
    # r_square = Column(DOUBLE(asdecimal=False)) #决定系数R方
    # sharpe = Column(DOUBLE(asdecimal=False)) #夏普率
    # tm_coef = Column(DOUBLE(asdecimal=False)) #风险收益指标
    

class FundScore(Base):
    '''基金评分'''

    __tablename__ = 'fund_score'
    id = Column(Integer, primary_key=True)

    fund_id = Column(CHAR(10)) #基金ID
    datetime =  Column(DATE) # 日期
    score = Column(DOUBLE(asdecimal=False)) #评分
    tag_name = Column(CHAR(64)) #基金类别
    tag_type = Column(Integer)#	1 '大类资产评分'
    tag_method = Column(CHAR(40)) #打标签版本号
    is_full = Column(Integer)#	1 true
    update_time	=  Column(DATE) # 更新时间
    score_method = Column(Integer)
    desc_name = Column(CHAR(64))




import dataclasses
import copy
import datetime
import numpy as np

@dataclasses.dataclass
class AssetDict:
    # 战略配置参数
    hs300: float=0.0
    csi500: float=0.0
    gem: float=0.0
    sp500rmb: float=0.0
    national_debt: float=0.0
    gold: float=0.0
    cash: float=0.0
    credit_debt: float=0.0

    def __repr__(self):
        s = f'<{self.__class__.__name__}'
        for k, v in self.__dict__.items():
            s += f' {k}={v}'
        s += '>'
        return s

    def copy(self):
        return copy.deepcopy(self)

    def isnan(self, v):
        return v is None or np.isnan(v)

@dataclasses.dataclass
class AssetWeight(AssetDict):

    def __post_init__(self):
        self.rebalance()

    def rebalance(self):
        _sum = 0.0
        for k, v in self.__dict__.items():
            assert v >= 0, f'{k} has negative value'
            _sum += v
        assert _sum > 0, 'param sum value must be positive'
        # normalize, make sure change is correct
        for k, v in self.__dict__.items():
            self.__dict__[k] /= _sum
    
    def __repr__(self):
        s = f'<{self.__class__.__name__}'
        for k, v in self.__dict__.items():
            s += f' {k}={v*100}%'
        s += '>'
        return s

@dataclasses.dataclass
class AssetPrice(AssetDict):

    def __post_init__(self):
        for k, v in self.__dict__.items():
            assert v is None or np.isnan(v) or v >= 0, f'{k} has negative value: {v}'
        self.cash = 1 # cash price is always 1
    
    def __repr__(self):
        s = f'<{self.__class__.__name__}'
        for k, v in self.__dict__.items():
            s += f' {k}={v}'
        s += '>'
        return s

@dataclasses.dataclass
class AssetPosition(AssetDict):

    def __post_init__(self):
        for k, v in self.__dict__.items():
            assert v >= 0, f'{k} has negative value: {v}'
    
    def __repr__(self):
        s = f'<{self.__class__.__name__}'
        for k, v in self.__dict__.items():
            s += f' {k}={v}'
        s += '>'
        return s

@dataclasses.dataclass
class AssetValue(AssetDict):

    prices: AssetPrice=None
    positions: AssetPosition=None

    def __post_init__(self):
        for index_id, p in self.prices.__dict__.items():
            self.__dict__[index_id] = 0 if self.isnan(p) else p * self.positions.__dict__[index_id]
            # 不允许出现价格为空，但是仓位为正的情况
            assert not (self.isnan(p) and self.positions.__dict__[index_id] > 0), \
                f'price is nan but position is not zero... (index){index_id} (p){p} (pos){self.positions.__dict__[index_id]}'

    def sum(self):
        _sum = 0.0
        for k in AssetDict.__dataclass_fields__.keys():
            _sum += self.__dict__[k]
        return _sum

    def __repr__(self):
        s = f'<{self.__class__.__name__}'
        for k, v in self.__dict__.items():
            s += f' {k}={v}'
        s += '>'
        return s

    def get_weight(self):
        wgts = AssetWeight(cash=1)
        for index_id, p in self.prices.__dict__.items():
            wgts.__dict__[index_id] = self.__dict__[index_id]
        wgts.rebalance()
        return wgts
        

@dataclasses.dataclass
class AssetTrade:

    index_id: str
    is_buy: bool
    mark_price: float
    submit_date: datetime.date
    amount: float=None
    volume: float=None
    trade_price: float=None
    commission: float=0.0
    trade_date: datetime.date=None

    def __post_init__(self):
        assert (self.amount and not self.volume) or (not self.amount and self.volume), \
            f'amount and volume, one and only one is valid (index){self.index_id} (vol){self.volume} (amt){self.amount}'

@dataclasses.dataclass
class AssetTradeParam:

    MinCountAmtDiff: float=0.001 # 少于总资产千分之一的资产Amount差异，不计在内
    MinActionAmtDiff: float=0.03 # 如果没有资产的偏离达到这个比例，那就不调仓

@dataclasses.dataclass
class TAAParam:

    HighThreshold: float=0.9    # 超过多少算作进入估值偏高的区域
    LowThreshold: float=0.1     # 低于多少算作进入估值偏低的区域
    HighMinus:    float=0.05    # 估值偏高时应该调低的占比
    LowPlus:      float=0.05    # 估值偏低是应该调高的占比
    HighStop:     float=0.5     # 进入估值偏高区后，什么时候算作退出
    LowStop:      float=0.5     # 进入估值偏低区后，什么时候算作退出

@dataclasses.dataclass
class FundScoreParam:
    tag_type: int=1
    score_method: int=1
    is_full: int=1

    def __repr__(self):
        return f'<fund_score_param tag_type={self.tag_type} score_method={self.score_method} is_full={self.is_full}'

class FundScoreHelper:
    ASSET_DICT = {
        '利率债' : 'national_debt',
        '美股大盘': 'sp500rmb',
        '德国大盘': 'dax30rmb',
        '信用债': 'credit_debt',
        '房地产': 'real_state',
        'A股大盘': 'hs300',
        '原油': 'oil',
        '黄金': 'gold',
        '创业板': 'gem',
        '日本大盘': 'n225rmb',
        'A股中盘': 'csi500',
    }
    TACTIC_ASSET_CONVERT = {
        '000300.XSHG': 'hs300',
        '000905.XSHG': 'csi500'
    }

    @staticmethod
    def keys():
        return FundScoreHelper.ASSET_DICT.keys()

    @staticmethod
    def get(index_id):
        for k, v in FundScoreHelper.ASSET_DICT.items():
            if v == index_id:
                return k
        return None
    
    @staticmethod
    def parse(tag_name):
        return FundScoreHelper.ASSET_DICT.get(tag_name)

if __name__ == '__main__':
    a = AssetWeight(hs300=0.1, csi500=0.1, cash=0.8)
    print(a)


from .rq_raw_data_downloader import RqRawDataDownloader
from .web_raw_data_downloader import WebRawDataDownloader

class RawDataDownloader(object):
    def __init__(self, rq_license):
        self.rq_downloader = RqRawDataDownloader(rq_license)
        self.web_downloader = WebRawDataDownloader()

    def download(self, start_date, end_date):
        self.rq_downloader.download_all(start_date, end_date)
        self.web_downloader.download_all(start_date, end_date)
        


import pandas as pd
from python.fund.db.database import DerivedDatabaseConnector, BasicDatabaseConnector
from python.fund.db.view_models import IndexDailyCollection
from python.fund.db.basic_models import IndexInfo, IndexPrice
from python.fund.db.derived_models import IndexValuation, IndexReturn, IndexVolatility
from sqlalchemy import create_engine


def get_index_volatility():
    with DerivedDatabaseConnector().managed_session() as mn_session:
        try:
            latest_time = mn_session.query(IndexVolatility).order_by(IndexVolatility.datetime.desc()).limit(1).one_or_none()
            latest_time = latest_time.datetime
            print(latest_time)
            query = mn_session.query(
                IndexVolatility.index_id,
                IndexVolatility.datetime,
                IndexVolatility.w1_vol,
                IndexVolatility.m1_vol,
                IndexVolatility.m3_vol,
                IndexVolatility.m6_vol,
                IndexVolatility.y1_vol,
                IndexVolatility.y3_vol,
                IndexVolatility.y5_vol,
                IndexVolatility.y10_vol,
                IndexVolatility.this_y_vol,
                IndexVolatility.cumulative_vol,
            ).filter(IndexVolatility.datetime==latest_time)
            df = pd.read_sql(query.statement, query.session.bind)
            df = df.rename(columns={'datetime': 'vol_datetime'})
            df = df.set_index('index_id')
            return df
        except Exception as e:
            print('Failed get_index_volatility <err_msg> {}'.format(e))

def get_index_valuation():
    with DerivedDatabaseConnector().managed_session() as mn_session:
        try:
            latest_time = mn_session.query(IndexValuation).order_by(IndexValuation.datetime.desc()).limit(1).one_or_none()
            latest_time = latest_time.datetime
            print(latest_time)
            query = mn_session.query(
                IndexValuation.index_id,
                IndexValuation.pb_mrq,
                IndexValuation.pe_ttm,
                IndexValuation.peg_ttm,
                IndexValuation.roe_ttm,
                IndexValuation.dy_ttm,
                IndexValuation.pe_pct,
                IndexValuation.pb_pct,
                IndexValuation.val_score,
                IndexValuation.datetime,
            ).filter(IndexValuation.datetime==latest_time)
            df = pd.read_sql(query.statement, query.session.bind)
            df = df.set_index('index_id')
            return df

        except Exception as e:
            print('Failed get_index_valuation <err_msg> {}'.format(e))

def get_index_return():
    with DerivedDatabaseConnector().managed_session() as mn_session:
        try:
            latest_time = mn_session.query(IndexReturn).order_by(IndexReturn.datetime.desc()).limit(1).one_or_none()
            latest_time = latest_time.datetime
            print(latest_time)
            query = mn_session.query(
                IndexReturn.index_id,
                IndexReturn.datetime,
                IndexReturn.w1_ret,
                IndexReturn.m1_ret,
                IndexReturn.m3_ret,
                IndexReturn.m6_ret,
                IndexReturn.y1_ret,
                IndexReturn.y3_ret,
                IndexReturn.y5_ret,
                IndexReturn.y10_ret,
                IndexReturn.this_y_ret,
                IndexReturn.cumulative_ret,
            ).filter(IndexReturn.datetime==latest_time)
            df = pd.read_sql(query.statement, query.session.bind)
            df = df.rename(columns={'datetime': 'ret_datetime'})
            df = df.set_index('index_id')
            return df

        except Exception as e:
            print('Failed get_index_return<err_msg> {}'.format(e))

def get_index_info():
    with BasicDatabaseConnector().managed_session() as mn_session:
        try:
            query = mn_session.query(
                IndexInfo.index_id,
                IndexInfo.order_book_id,
                IndexInfo.industry_tag,
                IndexInfo.tag_method,
                IndexInfo.desc_name,
            )
            df = pd.read_sql(query.statement, query.session.bind)
            df = df.set_index('index_id')
            return df

        except Exception as e:
            print('Failed get_index_return<err_msg> {}'.format(e))

def get_index_price():
    with BasicDatabaseConnector().managed_session() as mn_session:
        try:
            latest_time = mn_session.query(IndexPrice).order_by(IndexPrice.datetime.desc()).limit(1).one_or_none()
            latest_time = latest_time.datetime
            print(latest_time)
            query = mn_session.query(
                IndexPrice.index_id,
                IndexPrice.datetime,
                IndexPrice.volume,
                IndexPrice.low,
                IndexPrice.close,
                IndexPrice.high,
                IndexPrice.open,
                IndexPrice.total_turnover,
                IndexPrice.ret,
            ).filter(IndexPrice.datetime==latest_time)
            df = pd.read_sql(query.statement, query.session.bind)
            df = df.rename(columns={'datetime': 'price_datetime'})
            df = df.set_index('index_id')
            return df

        except Exception as e:
            print('Failed get_index_return<err_msg> {}'.format(e))

def append_data(table_name, data_append_directly_data_df):
    engine = create_engine(
            'mysql+pymysql://wangjun:Wangjun123456@fundinfo.cq1tbd5lkqzo.rds.cn-northwest-1.amazonaws.com.cn:3306/view')

    if not data_append_directly_data_df.empty:
        data_append_directly_data_df.to_sql(table_name, engine, index = False, if_exists = 'append')
        print('新数据已插入')
    else:
        print('没有需要插入的新数据')

def collection_daily_index():

    print('1、 load data...')
    info = get_index_info()
    vol = get_index_volatility()
    ret = get_index_return()
    val = get_index_valuation()
    price = get_index_price()

    print('2、 concat data...')
    df = pd.concat([info, vol, ret, val, price], axis=1, sort=False)
    df.index.name = 'index_id'
    df = df.reset_index()

    print('3、 special option...')

    df['order_book_id'] = df['order_book_id'].apply(lambda x: x.split('.')[0] if x != 'not_available' else None)
    df['industry_tag'] = df['industry_tag'].apply(lambda x: x if x != 'not_available' else None)

    append_data(IndexDailyCollection.__tablename__, df)

    print(df)


if __name__ == '__main__':
    collection_daily_index()




from .config import Configurator
from .struct import WxRawData, WxSeqItem, WxMsgItem
from .processor import IfaProcessor
import time, json

import sys
sys.path.append('/shared/build_surfing/src')
import libwxwork

class WxworkRunner(object):

    def __init__(self, seq=None):
        self.seq = seq
        corp_id = Configurator().config['wxwork']['corp_id']
        secret_key = Configurator().config['wxwork']['secret_key']
        private_key = Configurator().config['wxwork']['private_key']
        self.wrapper = libwxwork.WxworkWrapper(corp_id, secret_key, private_key, -1)
        self.processors = []

    def register(self, processor: IfaProcessor):
        self.processors.append(processor)

    def run(self):
        print(self.wrapper.init())
        self.seq = self.seq or 0
        while True:
            time.sleep(1)
            res = self.wrapper.get(self.seq, 500)
            #try:
            data = WxRawData(**json.loads(res))
            for item in data.chatdata:
                self.process(item)
            self.seq = data.max_seq or self.seq
            print(data)

    def process(self, item: WxSeqItem):
        decrypted = None
        try:
            decrypted_str = self.wrapper.decode(item.encrypt_random_key, item.encrypt_chat_msg)
            decrypted = json.loads(decrypted_str)
        except:
            print('!' * 10)
            print(f'error: {decrypted_str}')
            print('!' * 10)
        if decrypted:
            msg = WxMsgItem(decrypted)
            for pr in self.processors:
                self.processors.process(msg)
            print(msg)

if __name__ == "__main__":
    if len(sys.argv) > 1:
        config_path = sys.argv[1]
        print(f'config path is set: {config_path}')
        Configurator(config_path).config_path = config_path
        if len(sys.argv) > 2:
            seq = int(sys.argv[2])
    wmp = WxworkRunner(seq)
    wmp.run()


import pandas as pd
import numpy as np
from datetime import datetime
from datetime import date
import json
from fund.db.database import BasicDatabaseConnector,DerivedDatabaseConnector,RawDatabaseConnector
from fund.data_api.raw import get_raw_cm_index_price_df, get_cxindex_index_price_df, get_yahoo_index_price_df, get_rq_index_price_df,get_rq_index_weight,get_rq_stock_valuation,get_stock_fin_fac
from fund.data_api.basic import get_index_info, get_trading_day_list, get_index_price, get_fund_ret,get_fund_info,get_fund_nav
import rqdatac as rq
rq.init()

'''
当前只算来 hs300 csi500 pe 百分位
其他算法可以参考米筐公式 下面所有数可以通过rq.get_factor获得

pb_ratio_lf = market_cap_3 / equity_parent_company_mrq_0
pe_ratio_ttm = market_cap_3 / net_profit_parent_company_ttm_0
peg_ratio_ttm = pe_ratio_ttm / ((net_profit_ttm_0 - net_profit_ttm_1) / net_profit_ttm_1)
dividend_yield_ttm = dividend_per_share / close_price
return_on_equity_ttm = net_profit_parent_company_ttm_0 * 2 / (ma_0 + equity_parent_company_ttm_1)
第一遍算错，用了未来函数，第二次在第一次基础上更改
'''


index_df = get_index_info()
index_df = index_df[index_df['index_id'].isin(['hs300','csi500'])]
trade_list = rq.get_trading_dates(date(2005,1,1), date(2020,3,30))
trade_list = [ _ for _ in trade_list if _ > date(2005,4,15)]
result = {}
for index_o in index_df.order_book_id.tolist():
    result[index_o] = {}
    for d in trade_list:
        res_i = rq.index_weights(index_o, d)
        result[index_o][d] = res_i
stock_list = []
for index_i in result:
    print(index_i)
    for d in result[index_i]:
        stock_list.extend(result[index_i][d].index.tolist())
stock_list = list(set(stock_list))  
stock_cap = rq.get_factor(stock_list, 'market_cap_3',trade_list[0],trade_list[-1])
stock_profit = rq.get_factor(stock_list, 'net_profit_parent_company_ttm_0',trade_list[0],trade_list[-1])

res = []
for index_i in result:
    for d in result[index_i]:
        stock_list_i = result[index_i][d].index.tolist()
        p = stock_cap.loc[d,stock_list_i].sum() 
        e = stock_profit.loc[d,stock_list_i].sum() 
        res.append({
            'index_id' : index_i,
            'datetime' : d,
            'pe_ratio_ttm'    : p / e ,
        })    
index_val = pd.DataFrame(res)
res = []
for index_i in index_df['order_book_id'].to_list():
    index_val_i  = index_val[index_val['index_id'] == index_i]
    index_val_i.loc[:,'pe_pct'] = index_val_i['pe_ratio_ttm'].rank(pct=True).copy()
    res.append(index_val_i)
index_df = pd.concat(res)  

index_list = ['000905.XSHG','000300.XSHG']
pctrank = lambda x: pd.Series(x).rank(pct=True).iloc[-1]
res = []
for index_i in index_list:
    df_index = index_df[index_df['index_id'] == index_i]
    df_index.loc[:,'pe_pct'] = df_index.set_index('datetime')[['pe_ratio_ttm']].rolling(242*5).apply(pctrank, raw=True).values
    res.append(df_index)

df = pd.concat(res).drop(['id'], axis = 1)

df.to_sql('index_valpct', RawDatabaseConnector().get_engine(), index=False, if_exists='append')


from ...data.struct import AssetWeight, AssetPrice, AssetPosition, AssetValue
from ...data.struct import AssetTrade, AssetTradeParam
from . import Helper

class AssetTrader(Helper):

    def __init__(self, param: AssetTradeParam=None):
        self.param = param or AssetTradeParam()

    def calc_trade(self, dt, cur_position: AssetPosition, cur_price: AssetPrice, target_allocation: AssetWeight):

        cur_mv = AssetValue(prices=cur_price, positions=cur_position)
        new_position = cur_position.copy()
        new_position.cash = 0
        tot_mv = cur_mv.sum()

        trades = []
        launch_trade = False
        for index_id, target_weight in target_allocation.__dict__.items():
            if index_id != 'cash':
                target_amt = tot_mv * target_weight
                p = cur_price.__dict__[index_id]
                cur_amt = cur_mv.__dict__[index_id]
                if abs(target_amt - cur_amt) > tot_mv * self.param.MinCountAmtDiff:
                    volume = abs(target_amt - cur_amt) / p
                    is_buy = target_amt > cur_amt
                    trades.append(AssetTrade(
                        index_id=index_id, 
                        mark_price=p, 
                        volume=volume, 
                        is_buy=is_buy,
                        submit_date=dt
                    ))
                    new_position.__dict__[index_id] += volume if is_buy else -volume
                launch_trade = launch_trade or abs(target_amt - cur_amt) > tot_mv * self.param.MinActionAmtDiff
        new_mv = AssetValue(prices=cur_price, positions=new_position)
        new_position.cash = tot_mv - new_mv.sum()
        
        if not launch_trade:
            return cur_position, None
        else:
            print()
            print('----', dt)
            print('old', cur_position)
            print('new', new_position)
            print('old_mv', cur_mv)
            print('new_mv', new_mv)
            print()
            return new_position, trades

    def finalize_trade(self, dt, trades: list, t1_price: AssetPrice, bt_position: AssetPosition):
        pendings = []
        if trades is None or len(trades) == 0:
            return pendings
        # TODO: if some trades needs more time
        for trd in trades:
            trd.trade_price = t1_price.__dict__[trd.index_id]
            trd.trade_date = dt
            if not trd.amount:
                trd.amount = trd.volume * trd.trade_price
            if not trd.volume:
                trd.volume = trd.amount / trd.trade_price
            # TODO: commision calculate
            #trd.commission = ?
            # update position
            bt_position.__dict__[trd.index_id] += trd.volume if trd.is_buy else -trd.volume
            bt_position.cash += (-trd.amount if trd.is_buy else trd.amount) - trd.commission

        print()
        print('trade: ', trades)
        print()

        return pendings


import numpy as np
import pandas as pd
import datetime
from datetime import date
from pyfinance.ols import PandasRollingOLS
from fund.db.database import BasicDatabaseConnector,DerivedDatabaseConnector
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error, r2_score
from fund.data_api.basic import get_index_info, get_trading_day_list, get_index_price, get_fund_ret,get_fund_info,get_fund_nav
from fund.data_api.derived import get_fund_nav_adjusted_nv, get_index_benchmark_data,get_fund_list,get_fund_fee,get_fund_asset,get_fund_indicator


class SWElm:

    def __init__(self,alpha=0,beta=0,fee_rate=0,track_err=0,tag_method=None):
        self.alpha = alpha
        self.beta = beta
        self.fee_rate = fee_rate
        self.track_err = track_err
        self.tag_method = tag_method

def get_score(r,m):
    if m == 1: # 如果当前该资产只有一只基金，选吧
        return 1
    else:
        alv = (r.alpha - avg.alpha) / std.alpha
        bev = (r.beta   -avg.beta) / std.beta
        trv = (r.track_err - avg.track_err) /std.track_err
        frv = (r.fee_rate  - avg.fee_rate) / std.fee_rate
        m = score_dict[r.asset_type]
        result = m.alpha * alv + m.beta * abs(1-bev) + m.track_err * trv + m.fee_rate * frv
    return result        
        
def check_fund_number(fund_indicator):
    d_l= len(list(set(fund_indicator['datetime'].tolist())))
    tag_list  = list(set(fund_indicator['asset_type'].tolist()))
    for t in tag_list:
        l = fund_indicator[fund_indicator['asset_type'] == t].shape[0]
        print(t, l, l/d_l)    

score_dict = {}
tag_method = 'm0'

score_dict['A股大盘'] = SWElm(alpha=0.6, beta=-0.2, fee_rate = -0.2,tag_method=tag_method)
score_dict['A股中盘'] = SWElm(alpha=0.6, beta=-0.2, fee_rate = -0.2,tag_method=tag_method)
score_dict['创业板'] = SWElm(alpha=0.6, beta=-0.2, fee_rate = -0.2,tag_method=tag_method)
score_dict['美股大盘'] = SWElm(fee_rate = -0.2,track_err=-0.8,tag_method=tag_method)
score_dict['德国大盘'] = SWElm(fee_rate = -0.2,track_err=-0.8,tag_method=tag_method)
score_dict['日本大盘'] = SWElm(fee_rate = -0.2,track_err=-0.8,tag_method=tag_method)
score_dict['利率债'] = SWElm(alpha=0.2, fee_rate = -0.2,track_err=-0.6,tag_method=tag_method)
score_dict['信用债'] = SWElm(alpha=0.2, fee_rate = -0.2,track_err=-0.8,tag_method=tag_method)
score_dict['黄金'] = SWElm(alpha=0.2, fee_rate = -0.2,track_err=-0.8,tag_method=tag_method)
score_dict['原油'] = SWElm(alpha=0.2, fee_rate = -0.2,track_err=-0.8,tag_method=tag_method)
score_dict['房地产'] = SWElm(alpha=0.2, fee_rate = -0.2,track_err=-0.8,tag_method=tag_method)


fund_info = get_fund_info()
fund_info = fund_info[fund_info['asset_type'] != 'null']
asset_type = {i['fund_id'] : i['asset_type'] for i in fund_info.to_dict('records')}
fund_list = fund_info['fund_id'].to_list()
fund_indicator = get_fund_indicator(fund_list)
fund_indicator['asset_type'] = fund_indicator['fund_id'].map(lambda x : asset_type[x] )
fund_indicator['fee_rate'] = fund_indicator['fee_rate']

tag_list = list(score_dict.keys())
date_list = sorted(list(set(fund_indicator['datetime'].tolist())))
result = []
# 各资产日均基金数 
# 信用债 A股中盘 创业板 利率债 美股大盘 原油 德国大盘 A股大盘 房地产 黄金
#  35   19       8     12    3      6     1     40     23    11  
# 对于个数少的类别，筛选条件不严格

for d in date_list: 
    fid = fund_indicator[fund_indicator['datetime'] ==  d].copy()
    for t in tag_list:
        fitd = fid[fid['asset_type'] == t].copy()
        if fitd.shape[0] == 0:
            continue
        if t in ['创业板', '利率债', '美股大盘', '原油', '德国大盘', '黄金']:
            pass 
        elif t in ['信用债','A股中盘', 'A股大盘','房地产']:
            avg = fitd.mean()
            std = fitd.std()
            con1 = fitd.beta > (avg - std).beta
            con2 = fitd.beta < (avg + std).beta
            con3 = fitd.track_err > (avg - std).track_err
            fitd = fitd[con1 & con2 & con3].copy()
            if fitd.shape[0] < 5:
                continue
        avg = fitd.mean()
        std = fitd.std()
        result_i = []
        num = fitd.shape[0] 
        for i,r in fitd.iterrows():
            res  = get_score(r, num)
            r = r.to_dict()
            r['score'] = res
            result_i.append(r)
        df_i = pd.DataFrame(result_i)
        # TODO if df.shape[0] == 1: if df.shape[0] df.std = nan, np.std(df) = 0 
        # 
        avg = df_i['score'].mean()  
        std = df_i['score'].std()  
        con = (avg - std) < df_i['score']
        df_i = df_i[con]        
        df_i['score']  = (df_i['score']  - df_i['score'].min() )/ (df_i['score'].max() - df_i['score'].min())
        result.append(df_i.copy())

score_df = pd.concat(result)

fd_dic = {r.fund_id : {'start': r.start_date, 'end': r.end_date} for i, r in fund_info.iterrows()}
is_full = [r.datetime < fd_dic[r.fund_id]['end'] and r.datetime > fd_dic[r.fund_id]['start'] for i, r in score_df.iterrows()] 
score_df['is_full'] =  [int(_) for _ in is_full]
score_df = score_df.drop(['alpha','beta','fee_rate','track_err'] ,axis = 1)
name_dict = {
    'asset_type':'tag_name',
    'datetime':'datetime',
    'fund_id':'fund_id',
    'score':'score',
    'is_full':'is_full',
}
score_df = score_df.rename(columns=name_dict)
score_df.loc[:,'tag_type'] = 1 # 1 '大类资产评分'
score_df.loc[:,'tag_method'] = tag_method
score_df.loc[:,'score_method'] = 1 #'去一倍std beta 去一倍上界track_err 去一倍score score 0，1化 不同资产根据基金数决定筛选条件'
score_df.loc[:,'update_time'] = datetime.datetime.now()
score_df = score_df.replace(np.inf, np.nan)
score_df = score_df.replace(-np.inf, np.nan)
score_df = score_df.dropna()        
name_dict = {r.fund_id : r.desc_name for i,r in fund_info.iterrows()}      
score_df['desc_name'] = score_df['fund_id'].map(lambda x : name_dict[x])
score_df.to_sql('fund_score', DerivedDatabaseConnector().get_engine(), index=False, if_exists='append') 


import os
import json
import platform

from .helper import Singleton

DEFAULT_CONFIG_PATH = '/shared/surfing/config.json'
CURRENT_PLATFORM = platform.system()

class SysSettings(object):

    def __init__(self, sys_json):
        self.log_dir = sys_json.get('log_dir', '/shared/surfing/log/')
        self.temp_dir = sys_json.get('temp_dir', '/shared/surfing/temp/')  # 下载csv路径

        if CURRENT_PLATFORM == 'Darwin' or CURRENT_PLATFORM == 'Windows':
            curr_file_dir = os.path.split(os.path.realpath(__file__))[0]
            self.log_dir = curr_file_dir + '/../log/'
            self.temp_dir = curr_file_dir + '/../temp/'

        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)

        if not os.path.exists(self.temp_dir):
            os.makedirs(self.temp_dir)

    def get_temp_file_dir(self):
        return self.temp_dir

    def get_log_dir(self):
        return self.log_dir


class DbSettings(object):
    '''
    Parse database connection settings in config file
    '''
    def __init__(self, db_json):
        assert 'host' in db_json, 'could not find host in db settings'
        assert 'port' in db_json, 'could not find port in db settings'
        assert 'username' in db_json, 'could not find username in db settings'
        assert 'password' in db_json, 'could not find password in db settings'
        assert 'db' in db_json, 'could not find db in db settings'

        self.host = db_json['host']
        self.port = db_json['port']
        self.username = db_json['username']
        self.password = db_json['password']
        self.db = db_json['db']

    def to_conn_str(self):
        return 'mysql+pymysql://{}:{}@{}:{}/{}'.format(
            self.username, self.password, self.host, self.port, self.db
        )

    def __str__(self):
        return '<DB: {}@{}:{}/{}>'.format(self.username, self.host, self.port, self.db)


class HostSettings(object):

    def __init__(self, host_json):
        assert 'host' in host_json, 'could not find host in host_info settings'
        assert 'port' in host_json, 'could not find port in host_info settings'

        self.host = host_json['host']
        self.port = host_json['port']

    def to_dict(self):
        return {
            'host': self.host,
            'port': self.port,
        }

    def __str__(self):
        return '<HOSTING: {}@{}>'.format(self.host, self.port)


class SurfingConfigurator(metaclass=Singleton):

    def __init__(self, config_path=DEFAULT_CONFIG_PATH):
        self.config_path = config_path
        if CURRENT_PLATFORM == 'Darwin' or CURRENT_PLATFORM == 'Windows':
            curr_file_dir = os.path.split(os.path.realpath(__file__))[0]
            self.config_path = os.path.join(curr_file_dir, '../../../etc/config.json')
        self.config = json.load(open(self.config_path))

    def get_host_settings(self):
        assert 'hosting' in self.config, 'could not find hosting section in config'
        settings = HostSettings(self.config['hosting'])
        return settings

    def get_sys_settings(self):
        assert 'sys' in self.config, 'could not find sys section in config'
        settings = SysSettings(self.config['sys'])
        return settings

    def _get_db_settings(self, db_mode=None, key=None):
        assert key in self.config, 'could not find {} section in config'.format(key)
        db_info = db_mode or self.config[key].get('mode', 'developpment')
        assert db_info in self.config[key], 'could not find "{}" in config section'.format(db_info)
        db_section = self.config[key][db_info]
        settings = DbSettings(db_section)
        return settings

    def get_db_settings(self, db_mode=None, key=None):
        return self._get_db_settings(db_mode=db_mode, key=key)



import time
import signal
import datetime
import threading

from fund.util.timer import Timer
from fund.util.logger import SurfingLogger

from .common import DataType
from .data_loader import DataLoader
from .data_processor import DataProcessor
from .data_writer import DataWriter
from .dp_modules.pe_module import PeDailyModule

class DataEngine_test(object):
    '''
    Data engine consists of 3 parts, and schedules the data and work:
    1. DataLoader: load data from DB to memory
    2. DataProcessor: do the calculating using data in memory
    3. DataWriter: write data from memory to database
    DataLoader -> DataProcessor -> DataWriter
    '''

    def __init__(self):
        '''
        Initialization
        '''
        self._logger = SurfingLogger.get_logger(type(self).__name__)

        # Last data load timestamp, None if never
        self._data = {}

        # Data loader used to load database
        self._data_loader = DataLoader()

        # Data processor
        self._data_processor = DataProcessor()

        # Attach data processing modules to data_processor
        self._data_processor.attach(0, PeDailyModule())
        # self._data_processor.attach(0, TransactionModule())
        # self._data_processor.attach(0, UsdtPriceModule())
        # self._data_processor.attach(1, PerformanceModule())
        # self._data_processor.attach(2, RiskModule())
        # self._data_processor.attach(3, RiskEventModule())

        # Data writer
        self._data_writer = DataWriter()

        # Interrupt signal
        self._received_signal = None

    def _signal_handler(self, signum=None, frame=None):
        '''
        Handle interrupt signals
        '''
        if self._received_signal is None:
            self.stop(signum=signum)

    def start(self):
        '''
        Start data engine
        '''
        # Register Interrupt handler
        signal.signal(signal.SIGTERM, self._signal_handler)
        signal.signal(signal.SIGINT, self._signal_handler)

        # Start data engine
        time_interval_seconds = 60*60*24
        th = threading.Thread(
            target=self._run_periodic_task,
            args=[time_interval_seconds])
        th.setDaemon(True)
        th.start()

        while True:
            time.sleep(0.1)
            if self._to_stop():
                break

    def stop(self, signum=signal.SIGTERM):
        '''
        Stop data engine
        '''
        self._logger.info('Stopping, please wait...')
        self._received_signal = signum
        self._on_stop()

    def _on_stop(self):
        '''
        Called before stopping, could be used to save data
        '''
        self._logger.warn('Data engine is going to stop!')

    def _to_stop(self):
        '''
        Whether data engine needs to be stopped
        '''
        return self._received_signal is not None

    def _run_periodic_task(self, time_interval_seconds):
        '''
        Run periodic task
        :param time_interval_seconds: time interval of periodic tasks
        '''
        if time_interval_seconds is None or time_interval_seconds < 0:
            self._logger.error(f'Invalid time interval <time_interval_seconds> {time_interval_seconds}')
            self.stop()

        last_start = None
        while not self._to_stop():
            now = datetime.datetime.now()
            today = datetime.datetime.combine(now.date(), datetime.time())
            delta = (now - today).seconds
            now = today + datetime.timedelta(seconds=delta)
            residual = delta % time_interval_seconds

            next_start = None
            if last_start is None:
                if residual != 0:
                    time_to_sleep = time_interval_seconds - residual
                    if time_to_sleep > 2:
                        time_to_sleep /= 2
                    time.sleep(time_to_sleep)
                    continue
                else:
                    next_start = now
            else:
                next_start = last_start + datetime.timedelta(seconds=time_interval_seconds)
            self._logger.info(f'now: {now}, next_start: {next_start}')
            if now >= next_start:
                try:
                    calculate_daily = delta < time_interval_seconds
                    if last_start is not None:
                        last_start_nano = Timer.time_to_nano(last_start.timestamp())
                    else:
                        last_start_nano = None

                    self._run_task(
                        last_start_nano=last_start_nano,
                        curr_nano=Timer.time_to_nano(next_start.timestamp()),
                        calculate_daily=calculate_daily
                    )
                except Exception as e:
                    self._logger.error(f'Periodic task execution error <err_msg>{e}')
                last_start = next_start
            else:
                time_to_sleep = time_interval_seconds - residual
                if time_to_sleep > 2:
                    time_to_sleep /= 2
                time.sleep(time_to_sleep)
                continue

    def _run_task(self, last_start_nano=None, curr_nano=None, calculate_daily=None):
        self._logger.info('Start to run task')

        # Load initial data from database
        data_loaded = self._data_loader.load(start_time=last_start_nano, end_time=curr_nano, date=calculate_daily, internal_data=self._data)
        # TODO: merge data_loaded and self._data, not just assignment!
        self._data.update(data_loaded)
        # self._data[DataType.CALCULATE_DAILY] = calculate_daily
        self._logger.info('Start to process')
        # Run data processing
        status = self._data_processor.process(self._data)

        if not status:
            self._logger.error('Error in data_processor')

        # Write processed data back to database, and clear output data buffer
        data_writer_output = self._data_writer.write(self._data)
        self._data.update(data_writer_output)


if __name__ =='__main__':
    a = DataEngine_test()
    a._run_task(calculate_daily='20190201')


from fund.db.database import DerivedDatabaseConnector,DerivedTestDatabaseConnector
from fund.util.logger import SurfingLogger
from fund.db.models import IndustryPE
from fund.util.utility import Utility

from .common import DataType
import datetime


class DataWriter(object):
    '''
    Data writer, which writes data in memory back to database
    (Memory) -> DataWriter -> (DB)
    '''

    def __init__(self):
        '''
        Initialization
        '''
        self._logger = SurfingLogger.get_logger(type(self).__name__)

    def write(self, data):
        '''
        Main function of DataWriter, to be called by DataEngine.
        :param data: data to be written to DB: {DataType->data}
        '''
        # Write data to FoF Database
        self._logger.info('Start writing data to db')
        output = {}
        data_write_targets = [
            [DataType.UPDATE_INDUSTRY_PE, 'positions'],
            [DataType.UPDATE_STOCK_ANNUAL_NET_PROFIT, 'new stock annual net profit'],
        ]
        for data_type, data_name in data_write_targets:
            if data_type in data:
                output[data_type] = []
                items = data[data_type]
                if type(items) != list:
                    self._logger.error(f'Data {data_name} to written is not list')

                if len(items) == 0:
                    continue
                
                self._logger.info(f'Start writing {data_name} to db')
                if self._write_derived_result(items):
                    self._logger.info(f'Done writing {data_name} to db')
                else:
                    self._logger.error(f'Failed to write {data_name} to db')

        return output

    def _write_derived_result(self, items):
        '''
        Write processed data back to DB
        :return: True for success, and False for failure
        '''
        self._logger.info(f'data len: {len(items)}')
        with DerivedDatabaseConnector().managed_session() as derived_session:
            try:
                derived_session.bulk_save_objects(items)
                derived_session.commit()
            except Exception as e:
                derived_session.rollback()
                self._logger.error(f'Write error <err>{e}')
                return False
        return True
